// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"fmt"
	"os"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v4/tools"
)

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// listAccounts parses arguments to list accounts.
func listAccounts(_ *cobra.Command, _ []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(csrHash bool, certDetails bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	accounts, err := conn.AccountList()
	if err != nil {
		Logger.Error().Msgf("Cannot retrieve accounts: %s", err)
		return 1
	}

	if accountCount := len(accounts); accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		table := tablewriter.NewWriter(os.Stdout)
		headers := []string{"id", "name", "type", "status"}
		if csrHash {
			headers = append(headers, "CSR fingerprint")
		}
		if certDetails {
			headers = append(headers, "CSR", "Cert")
		}
		table.SetHeader(headers)
		table.SetBorder(false)
		table.SetHeaderLine(false)
		table.SetColumnSeparator(" ")

		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			values := []string{account.GetIDAsUUID().String(), account.Name, account.Type.String(), account.Status.String()}
			if csrHash {
				values = append(values, tools.GetCSRHash(account.CSR, true))
			}
			if certDetails {
				values = append(values, account.CSR, account.Cert)
			}
			table.Append(values)
		}
		table.Render()
	}
	return 0
}

func init() {
	accountCmd.AddCommand(accountListCmd)

	flags := accountListCmd.PersistentFlags()
	flags.Bool("cert-details", false, "Display cert details")
	must(viper.BindPFlag("cert-details", flags.Lookup("cert-details")))
	flags.Bool("csr-hash", true, "Display CSR fingerprints")
	must(viper.BindPFlag("csr-hash", flags.Lookup("csr-hash")))
}
