// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"os"

	"github.com/spf13/cobra"
	"golang.org/x/term"
	"xbus.io/go-xbus/v4/api/control"
)

var accountPasswordCmd = &cobra.Command{
	Use:   "password <identifier>",
	Short: "Set/change the password of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			if err := cmd.Usage(); err != nil {
				Logger.Error().Msgf("Failed to print usage: %s", err)
			}
			os.Exit(1)
		}
		var reset = cmd.Flag("reset").Value.String() == "true"
		os.Exit(accountPassword(args[0], reset))
	},
}

func accountPassword(identifier string, reset bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Error().Msgf("Could not find account %s: %s", identifier, err)
		return 1
	}

	var newpassword string

	if !reset {
		fd := int(os.Stdin.Fd())
		if !term.IsTerminal(fd) {
			Logger.Error().Msgf("Cannot read password from non-terminal")
			return 1
		}
		if outFd := int(os.Stdout.Fd()); !term.IsTerminal(outFd) {
			Logger.Error().Msgf("Cannot read password if stdin is a terminal and stdout is not")
			return 1
		}

		os.Stdout.WriteString("New password: ")
		b, err := term.ReadPassword(fd)
		os.Stdout.WriteString("\n")
		if err != nil {
			Logger.Error().Msgf("Failed to read password: %s", err)
			return 1
		}
		newpassword = string(b)

		os.Stdout.WriteString("Confirm password: ")
		b, err = term.ReadPassword(fd)
		os.Stdout.WriteString("\n")
		if err != nil {
			Logger.Error().Msgf("Failed to read password: %s", err)
			return 1
		}
		if newpassword != string(b) {
			Logger.Error().Msgf("Passwords mismatched. Please try again")
			return 1
		}
	}

	// TODO add a confirmation message
	if err := conn.Control.Account.PasswordSet(
		&control.AccountPasswordSetRequest{
			AccountID: account.ID,
			Password:  newpassword,
		},
	); err != nil {
		Logger.Error().Msgf("Failed: %s", err)
	}
	if reset {
		Logger.Info().Msgf("Reset %s password.", account.Name)
	} else {
		Logger.Info().Msgf("Updated %s password.", account.Name)
	}
	return 0
}

func init() {
	flags := accountPasswordCmd.Flags()
	flags.Bool("reset", false, "Reset the account password")

	accountCmd.AddCommand(accountPasswordCmd)
}
