// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"

	"github.com/spf13/cobra"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// actorRejectCmd represents the actor reject sub-command.
var actorRejectCmd = &cobra.Command{
	Use:   "reject",
	Short: "Reject an actor.",
	Long: "Reject a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: rejectActor,
}

// rejectActor parses arguments to reject a pending actor.
func rejectActor(cmd *cobra.Command, args []string) {
	var identifier string
	v := cmd.Flag("id")
	switch {
	case v.Changed:
		identifier = v.Value.String()
	case len(args) == 1:
		identifier = args[0]
	default:
		usage(cmd)
	}
	os.Exit(doRejectActor(identifier))
}

// doRejectActor rejects a pending actor request.
func doRejectActor(identifier string) int {
	if len(identifier) == 0 {
		Logger.Error().Msgf("Please provide an actor identifier.")
		return 1
	}

	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	actor, err := conn.ActorFind(identifier)
	if err != nil {
		if errors.Is(err, xbus.ErrNotFound) {
			Logger.Error().Msgf("actor not found")
		} else {
			Logger.Error().Msgf("could not find actor: %s", err)
		}
		return -1
	}

	// Send an "reject" request.
	repActor, err := conn.ActorReject(actor.GetIDAsUUID())
	if err != nil {
		Logger.Error().Msgf("Could not reject the actor %q: %s", identifier, err)
		return 1
	}
	if repActor.Status == api.Actor_PENDING {
		Logger.Error().Msgf("Could not reject the actor %q: Still pending.", identifier)
		return 1
	}
	Logger.Info().Msgf("Successfully rejected the actor %q (ID %q).", repActor.Name, identifier)
	return 0
}

func init() {
	actorCmd.AddCommand(actorRejectCmd)
}
