// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"errors"
	"os"
	"time"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
)

var envelopeGetCmd = &cobra.Command{
	Use:   "get <ID>",
	Short: "Retrieve a single envelope",
	Args: func(cmd *cobra.Command, args []string) error {
		if err := cobra.ExactArgs(1)(cmd, args); err != nil {
			return err
		}
		_, err := api.UUIDFromString(args[0])
		if err != nil {
			return errors.New("Invalid uuid")
		}
		return nil
	},
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		id := api.MustUUIDFromString(args[0])
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		if cmd.Flag("stream").Value.String() == "true" {
			if err := c.API.EnvelopeStorage.Retrieve(
				ctx, "account", c.ID().String(),
				&api.EnvelopePosition{EnvelopeID: id.ToBytes(), Start: true},
				func(ctx context.Context, fragment *api.EnvelopeFragmentAndPosition) {
					b, err := api.JSONMarshal(
						[]api.ItemEncoding{api.ItemJSON, api.ItemText, api.ItemBinary},
						fragment.GetFragment())
					if err != nil {
						Logger.Error().Msgf("%s", err)
					}
					os.Stdout.Write(b)
					os.Stdout.Write([]byte("\n"))
				},
			); err != nil {
				Logger.Error().Msgf("receiving: %s", err)
			}
		} else {
			receiver := c.RetrieveEnvelope(ctx, "account", c.ID().String(), id)
			env, err := receiver.Complete(5 * time.Second)
			if err != nil {
				Logger.Error().Msgf("receiving: %s", err)
				os.Exit(1)
			}
			b, err := api.JSONMarshal(
				[]api.ItemEncoding{api.ItemJSON, api.ItemText, api.ItemBinary},
				env)
			if err != nil {
				Logger.Error().Msgf("%s", err)
				os.Exit(1)
			}
			os.Stdout.Write(b)
		}
	},
}

func init() {
	// envelope get
	flags := envelopeGetCmd.PersistentFlags()
	flags.Bool("stream", false, "Store a streamed version of the envelope")
	envelopeCmd.AddCommand(envelopeGetCmd)
}
