// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"fmt"
	"os"

	"github.com/spf13/cobra"
)

var pipelineSaveCmd = &cobra.Command{
	Use:   "save <name or id> [version] <filepath>",
	Short: "Save a pipeline",
	Long:  "Save a pipeline definition (its graph)",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 2 || len(args) > 3 {
			usage(cmd)
		}
		var (
			identifier = args[0]
			version    string
			filename   string
		)
		if len(args) == 2 {
			filename = args[1]
		} else {
			version = args[1]
			filename = args[2]
		}
		if err := pipelineSave(identifier, version, filename); err != nil {
			Logger.Error().Msgf("Error saving pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

func pipelineSave(identifier, version, filename string) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	content, err := os.ReadFile(filename)
	if err != nil {
		return fmt.Errorf("cannot read pipeline definition file: %w", err)
	}

	info := pipelineInfo(identifier, version)

	info, _, validationMessages, err := c.PipelineSave(info, string(content))
	if len(validationMessages) != 0 {
		fmt.Println("Validation messages:")
		for _, msg := range validationMessages {
			fmt.Println("-", msg)
		}
	}

	if err != nil {
		return err
	}

	Logger.Info().Msgf("Successfully saved pipeline %s-%s graph", info.Name, info.Version)
	return nil
}

func init() {
	pipelineCmd.AddCommand(pipelineSaveCmd)
}
