// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"fmt"
	"os"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

var psSummaryCmd = &cobra.Command{
	Use:   "summary",
	Short: "Show a summary of the processes",
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{IncludeAllStatuses: true})
		if err != nil {
			Logger.Error().Msgf("Invalid filter: %s", err)
			return
		}

		var summaryEntries []*control.ProcessSummary_Entry

		ctx, cancel := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		defer cancel()

		if err := c.Control.Process.QuerySummary(
			ctx, &control.ProcessQueryRequest{Filter: filter},
			func(ctx context.Context, r *control.ProcessSummary) {
				summaryEntries = append(summaryEntries, r.Entries...)
			},
		); err != nil {
			Logger.Error().Msgf("error getting summary: %s", err)
			os.Exit(1)
		}

		pipelineList, err := c.PipelineQuery("")
		if err != nil {
			Logger.Error().Msgf("%s", err)
			return
		}
		actorList, err := c.ActorList(nil)
		if err != nil {
			Logger.Error().Msgf("%s", err)
			return
		}
		pipelines := make(map[api.UUID]*api.PipelineInfo)
		for _, g := range pipelineList {
			pipelines[g.GetIDAsUUID()] = g
		}
		actors := make(map[api.UUID]*api.Actor)
		for _, a := range actorList {
			actors[a.GetIDAsUUID()] = a
		}

		table := tablewriter.NewWriter(os.Stdout)
		table.SetHeader([]string{"pipeline", "version", "emitter", "status", "count"})

		for _, entry := range summaryEntries {
			pipeline := entry.GetPipelineIDAsUUID().String()
			version := ""
			if g, ok := pipelines[entry.GetPipelineIDAsUUID()]; ok {
				pipeline = g.GetName()
				version = g.GetVersion()
			}
			actor := entry.GetEmitterIDAsUUID().String()
			if a, ok := actors[entry.GetEmitterIDAsUUID()]; ok {
				actor = a.GetName()
			}
			table.Append([]string{
				pipeline,
				version,
				actor,
				entry.Status.String(),
				fmt.Sprintf("%d", entry.Count),
			})
		}
		table.Render()
	},
}

func init() {
	psCmd.AddCommand(psSummaryCmd)
}
