// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
unencrypted "registration" channel. This xbusctl will only be allowed to send
Xbus actual commands once accepted on the Xbus side.
This command also outputs the command one has to run on the Xbus side to accept
the registration request.`,
	Run: func(cmd *cobra.Command, args []string) {
		os.Exit(doRegister(false, false))
	},
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister(renewKey, renewCert bool) int {
	xbusClient, err := GetClient(true)
	if err != nil {
		Logger.Error().Msgf("%s", err)
		return 1
	}

	if err := xbusClient.CreateKey(renewKey); err != nil {
		if !errors.Is(err, xbus.ErrAlreadyExists) {
			Logger.Error().Msgf("Cannot create a private key %s", err)
			return 1
		}
	}

	// TODO find the actual hostname or get one in the configuration
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR("localhost", "xbus", "xbusctl", renewCert); err != nil {
		if !errors.Is(err, xbus.ErrAlreadyExists) {
			Logger.Error().Msgf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Error().Msgf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Info().Msgf("Successfully sent the registration request.")

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.Registration_PENDING:
		Logger.Info().Msgf(`The registration is pending approval. To approve from Xbus, run:
	xbusd account accept %q`,
			xbusClient.Name(),
		)
	case api.Registration_ACCEPTED:
		Logger.Info().Msgf(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.Registration_REJECTED:
		Logger.Info().Msgf(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

func init() {
	RootCmd.AddCommand(registerCmd)
}
