// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"os"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/services/control"
	"xbus.io/xbus/storage"
)

// Arguments of the "accept" command.
const (
	acceptCmdArgKey = iota
	acceptCmdArgCount
)

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <id|name|fingerprint>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	ctx := MakeCmdContext()
	if len(args) < acceptCmdArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	certDir := viper.GetString("cert-dir")

	expire, err := getExpire(cmd, "cert-expire")
	if err != nil {
		Logger.Error().Msgf("%s", err)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}
	os.Exit(doAcceptAccount(ctx, getAccountStorage(), args[acceptCmdArgKey], certDir, expire))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(
	ctx context.Context, accountStorage storage.AccountStorage, key string, certDir string, expire time.Time,
) int {
	// Load the servers's client CA in order to sign CSRs.
	clientCACert, clientCAKey, err := loadClientCA(certDir)
	if err != nil {
		Logger.Error().Msgf("Cannot load the client CA: %s", err)
		return 1
	}

	account, found := findAccount(ctx, accountStorage, key)
	if !found {
		return 1
	}

	// Got a match! Ensure the account is pending.
	if account.Status != api.Account_PENDING {
		Logger.Error().Msgf("The specified account (%s) is not pending a validation.",
			account.Name)
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := control.AccountAccept(ctx, accountStorage, account.GetIDAsUUID(), clientCACert, clientCAKey, expire)
	if err != nil {
		Logger.Error().Msgf("Could not accept the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Error().Msgf("Could not accept the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Info().Msgf("Successfully accepted the account %q.", repAccount.Name)
	return 0
}

// loadClientCA loads the server CA made for client connections and provides the cert & the private
// key, so account signing functions, in particular, can use it.
// TODO This should be elsewhere.
func loadClientCA(certDir string) (
	clientCACert *x509.Certificate, clientCAKey interface{}, err error,
) {
	clientCAInfo, err := tls.LoadX509KeyPair(certDir+"/ClientCA.crt", certDir+"/ClientCA.key")
	if err != nil {
		return nil, nil, err
	}
	clientCACert, err = x509.ParseCertificate(clientCAInfo.Certificate[0])
	if err != nil {
		return nil, nil, err
	}
	clientCAKey = clientCAInfo.PrivateKey
	return clientCACert, clientCAKey, nil
}

func init() {
	accountCmd.AddCommand(accountAcceptCmd)

	flags := accountAcceptCmd.Flags()
	flags.String("cert-expire", "1y", "Date of duration for the account certificate to expire")
	must(viper.BindPFlag("account.cert-expire", flags.Lookup("cert-expire")))
}
