// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"xbus.io/xbus/storage"
)

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// listAccounts parses arguments to list accounts.
func listAccounts(_ *cobra.Command, _ []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(MakeCmdContext(), getAccountStorage(), csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(ctx context.Context, accountStorage storage.AccountStorage, csrHash bool, certDetails bool) int {
	accounts := accountStorage.List(ctx)

	if accountCount := len(accounts); accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			fmt.Println(fmtAccount(account, csrHash, certDetails))
		}
	}
	return 0
}

func init() {
	accountCmd.AddCommand(accountListCmd)
	flags := accountListCmd.PersistentFlags()
	flags.Bool("cert-details", false, "Display cert details")
	must(viper.BindPFlag("cert-details", flags.Lookup("cert-details")))
	flags.Bool("csr-hash", true, "Display CSR fingerprints")
	must(viper.BindPFlag("csr-hash", flags.Lookup("csr-hash")))
}
