// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/services/control"
	"xbus.io/xbus/storage"
)

// Arguments of the "reject" command.
const (
	rejectCmdArgKey = iota
	rejectCmdArgCount
)

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <id|name|fingerprint>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	os.Exit(doRejectAccount(MakeCmdContext(), getAccountStorage(), args[rejectCmdArgKey]))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(ctx context.Context, accountStorage storage.AccountStorage, key string) int {
	account, found := findAccount(ctx, accountStorage, key)
	if !found {
		return 1
	}

	// Got a match! Ensure the account is pending.
	if account.Status != api.Account_PENDING {
		Logger.Error().Msgf("The specified account (%s) is not pending a validation.",
			account.Name)
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := control.AccountReject(ctx, accountStorage, account.GetIDAsUUID())
	if err != nil {
		Logger.Error().Msgf("Could not reject the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Error().Msgf("Could not reject the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Info().Msgf("Successfully rejected the account %q.", repAccount.Name)
	return 0
}

func init() {
	accountCmd.AddCommand(accountRejectCmd)
}
