// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v4/tools"
)

// keyGenerateCmd represents the key generation sub-command.
var keyGenerateCmd = &cobra.Command{
	Use:   "generate",
	Short: "Generate private keys.",
	Long: `Generate private keys:

- ServerCA.key: For the server CA (the one clients connecting to Xbus will
  check).

- ClientCA.key: For the client CA (the one clients connecting to Xbus will have
  to have been signed by).`,
	Run: generateKeys,
}

// generateKeys parses arguments to generate private keys.
func generateKeys(_ *cobra.Command, _ []string) {
	certDir := viper.GetString("cert-dir")
	os.Exit(doGenerateKeys(certDir))
}

// doGenerateKeys generates private keys (see the command description for more information).
func doGenerateKeys(certDir string) int {
	generated := genKey(certDir+"/ServerCA.key") && genKey(certDir+"/ClientCA.key")
	if !generated {
		return 1
	}
	return 0
}

// genKey is a helper that creates a private key and saves it into a file.
func genKey(keyPath string) (generated bool) {
	privKeyData, err := tools.CreatePrivateKey()
	if err != nil {
		Logger.Error().Msgf("Error generating the key: %s", err)
		return false
	}

	if err := os.WriteFile(keyPath, privKeyData, 0644); err != nil {
		Logger.Error().Msgf("Error writing the key to %q: %s", keyPath, err)
		return false
	}

	Logger.Info().Msgf("Successfully generated %q.", keyPath)
	return true
}

func init() {
	keyCmd.AddCommand(keyGenerateCmd)
}
