// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package fullstack

import (
	"context"
	"fmt"
	"strconv"
	"sync"
	"sync/atomic"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"

	"xbus.io/xbus/services/envelopestorage"
	testutils "xbus.io/xbus/testutils"
)

/* This test has no meaning with the new way envelopes are retrieved by the
actor.
However, once we have a back-pressure on envelope reading, we can recycle
this test to make sure slow or very slow actors are ok.

func TestSlowConsumer(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`)

	var (
		itemCount     int
		fragmentCount int
	)

	csub, err := consumer.RawSubscribe(
		func(input string, envelope api.Envelope) (api.EnvelopeAck, error) {
			time.Sleep(50 * time.Millisecond)
			ack := api.EnvelopeAck{
				ID:     envelope.ID,
				Status: api.EnvelopeAck_ACCEPTED,
			}
			fragmentCount++
			itemCount += len(envelope.Events[0].Items)
			t.Log("Consumer received envelope -", itemCount, "items so far")
			if itemCount == 3000 {
				go consumer.ProcessingSuccess(envelope.Context, envelope.ID)
			}
			return ack, nil
		}, true)
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	var item = make([]byte, 1024)

	go func() {
		output := emitter.OpenOutput(api.ProcessingContext{}, "default", "evt1")
		for i := 0; i != 3000; i++ {
			output.AddItems("evt1", api.Item(item))
		}
		assert.Nil(t, output.Close())
		t.Log("Sent all the items")
	}()

	timeout := time.After(20 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.ProcessList[0].ID
	case <-timeout:
		ft.Logger.Error().Msgf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE, state.Status)
		assert.Empty(t, state.ProcessList[0].Errors)
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	case <-timeout:
		ft.Logger.Error().Msgf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	}
}
*/

func TestStreamedEnvelope(t *testing.T) {
	const emissionLoops = 20000
	ctx := testutils.GetTestContext(t)
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	_, worker := ft.CreateWorker("fw", []string{"forward"})
	wsub, err := worker.Subscribe(func(apc *xbus.ActorProcessingContext) error {
		worker.Log.Info().Msgf("Receiving envelope")
		input, err := apc.GetEnvelope(api.DefaultInput)
		if err != nil {
			t.Error(err)
			return err
		}

		output := envelope.MustNewOpenedEnvelope()
		go func() {
			if err := envelope.ForEachMessage(input, func(msg envelope.Message) error {
				outMsg := envelope.MustNewMessageWriter(msg.MsgType())
				if err := output.Add(outMsg); err != nil {
					return err
				}
				go func() {
					var count int
					assert.NoError(t, envelope.ForEachChunk(msg, func(b []byte) error {
						count++
						return outMsg.WriteChunk(b)
					}))
					fmt.Println("COUNTED", msg.MsgType(), count)
					assert.NoError(t, outMsg.Close())
				}()
				return nil
			}); err != nil {
				t.Error(err)
			}
			assert.NoError(t, output.Close())
		}()
		return apc.Send("default", output)
	})

	assert.NoError(t, err)
	defer testutils.DeferedAssertNoError(t, wsub.Unsubscribe)

	var (
		receivedEvt1Count int64
		receivedEvt2Count int64
	)

	csub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			input, err := apc.GetEnvelope(api.DefaultInput)
			if !assert.NoError(t, err) {
				return err
			}
			wg := sync.WaitGroup{}
			if err := envelope.ForEachMessage(input, func(msg envelope.Message) error {
				var counter *int64
				if msg.MsgType() == "evt1" {
					counter = &receivedEvt1Count
				} else {
					counter = &receivedEvt2Count
				}
				wg.Add(1)
				go func() {
					defer wg.Done()
					assert.NoError(t, envelope.ForEachChunk(msg, func(b []byte) error {
						c := atomic.AddInt64(counter, 1)
						if c < 6 {
							fmt.Println("RECEIVED", msg.MsgType(), string(b))
						}
						if msg.MsgType() == "evt1" && c >= 20000 {
							fmt.Println("RECEIVED", msg.MsgType(), string(b))
						}
						if msg.MsgType() == "evt2" && c >= 60000 {
							fmt.Println("RECEIVED", msg.MsgType(), string(b))
						}
						return nil
					}))
				}()
				return nil
			}); err != nil {
				return err
			}
			wg.Wait()

			if atomic.LoadInt64(&receivedEvt1Count) != emissionLoops ||
				atomic.LoadInt64(&receivedEvt2Count) != 3*emissionLoops {
				err := fmt.Errorf(
					"Expected %d evt1, %d evt2 - Received: %d evt1, %d evt2",
					emissionLoops, 3*emissionLoops,
					receivedEvt1Count, receivedEvt2Count)
				return err
			}
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer testutils.DeferedAssertNoError(t, csub.Unsubscribe)

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1,evt2]
  outputs: [default]
- id: relay
  type: worker
  actors: [fw]
  inputs: [default]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> relay
- relay.default -> target.default
`)

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	evt1 := envelope.MustNewMessageWriter("evt1")
	evt2 := envelope.MustNewMessageWriter("evt2")
	emission, err := emitter.CustomEmission(context.TODO(), envelope.MustNewEnvelope(evt1, evt2), false)
	require.NoError(t, err)
	for i := 0; i != emissionLoops; i++ {
		assert.NoError(t, evt1.WriteStringChunk(("evt1-item-" + strconv.Itoa(i))))
		assert.NoError(t, evt2.WriteStringChunk("evt2-0-item-"+strconv.Itoa(i)))
		assert.NoError(t, evt2.WriteStringChunk("evt2-1-item-"+strconv.Itoa(i)))
		assert.NoError(t, evt2.WriteStringChunk("evt2-2-item-"+strconv.Itoa(i)))
	}
	assert.NoError(t, evt1.Close())
	assert.NoError(t, evt2.Close())
	t.Log("Wrote the messages")

	timeout := time.After(180 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.GetProcessIDAsUUID()
	case <-timeout:
		ft.Logger.Error().Msgf("Process never started")
		t.FailNow()
	}

	select {
	case err := <-emission.Done():
		require.NoError(t, err)
		t.Log("emission is over")
	case <-timeout:
		t.Error("emission takes too long")
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE.String(), state.Status.String())
		assert.Empty(t, state.Errors)
		pState := ft.Server.ProcessStorage().GetState(ctx, processID)
		t.Log(pState.Fmt())
	case <-timeout:
		ft.Logger.Error().Msgf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(ctx, processID)
		t.Log(pState.Fmt())
	}
}

func switchDurationVar(v *time.Duration, nv time.Duration) func() {
	originalValue := *v
	*v = nv
	return func() {
		*v = originalValue
	}
}

func TestSlowEmitter(t *testing.T) {
	ctx := testutils.GetTestContext(t)
	defer switchDurationVar(
		&envelopestorage.EnvelopeStallDelay, 250*time.Millisecond)()
	defer switchDurationVar(
		&xbus.EnvelopeSenderDefaultDelay, 100*time.Millisecond)()

	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	cancel := make(chan struct{})

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	csub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			e, err := apc.CompleteEnvelope(api.DefaultInput, 5*time.Second)
			if err != nil {
				t.Error(err)
				return err
			}
			assert.Equal(t, 1, len(e.Events))
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer testutils.DeferedAssertNoError(t, csub.Unsubscribe)

	ft.InstallGraph(`
name: direct
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`)

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	go func() {
		evt1 := envelope.MustNewMessageWriter("evt1")
		emission, err := emitter.CustomEmission(
			context.TODO(), envelope.MustNewEnvelope(evt1), false)
		require.NoError(t, err)
		assert.Nil(t, evt1.WriteStringChunk("value"))
		assert.NoError(t, evt1.Close())
		time.Sleep(1000 * time.Millisecond)
		assert.Nil(t, emission.Wait())
	}()

	timeout := time.After(6 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.GetProcessIDAsUUID()
	case <-timeout:
		ft.Logger.Error().Msgf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE.String(), state.Status.String())
		assert.Empty(t, state.Errors)
		pState := ft.Server.ProcessStorage().GetState(ctx, processID)
		t.Log(pState.Fmt())
	case <-timeout:
		close(cancel)
		ft.Logger.Error().Msgf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(ctx, processID)
		t.Log(pState.Fmt())
	}
}
