// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package process

import (
	"context"
	"fmt"

	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/storage"
)

// NewManager creates a process.Manager
func NewManager(
	pipelineManager *pipeline.Manager,
	actorStorage storage.ActorStorage,
	processStorage storage.ProcessStorage,
	logger zerolog.Logger,
) *Manager {
	manager := Manager{
		PipelineManager: pipelineManager,
		ActorStorage:    actorStorage,
		ProcessStorage:  processStorage,
		Log: logger.With().
			Str("component", "process").
			Str("class", "process.Manager").
			Logger(),
	}
	return &manager
}

// Manager manage the processes
type Manager struct {
	PipelineManager *pipeline.Manager
	ActorStorage    storage.ActorStorage
	ProcessStorage  storage.ProcessStorage

	Log zerolog.Logger

	CurrentProcessList []*Process
}

// Close does nothing for now
func (m *Manager) Close() {
}

func (m *Manager) loadProcess(ctx context.Context, p storage.Process) (*Process, error) {
	var pipe *pipeline.Pipeline
	if !p.PipelineID.IsZero() {
		p1, exists := m.PipelineManager.GetPipeline(ctx, p.PipelineID)
		if !exists {
			return nil, fmt.Errorf("No such pipeline: %s", p.PipelineID)
		}
		pipe = p1
	}
	return &Process{
		ID:                p.ID,
		CreatedAt:         p.CreatedAt,
		ReplayOf:          p.ReplayOf,
		GroupID:           p.GroupID,
		actors:            m.ActorStorage,
		processStorage:    m.ProcessStorage,
		graphManager:      m.PipelineManager,
		manager:           m,
		TriggerEmitterID:  p.TriggerEmitterID,
		TriggerEnvelopeID: p.TriggerEnvelopeID,
		Pipeline:          pipe,
		SourceRef:         p.SourceOutput,
	}, nil
}

// LookupOrCreate returns the Process handling the given envelope or create
// if necessary
func (m *Manager) LookupOrCreate(
	ctx context.Context, emitter *api.Actor, output string, envelopeID api.UUID,
) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(ctx, emitter, envelopeID)
	if ok {
		return m.loadProcess(ctx, p)
	}

	p = storage.Process{
		TriggerEmitterID:  emitter.GetIDAsUUID(),
		TriggerEnvelopeID: envelopeID,
		SourceOutput:      storage.NodeOutputRef{Output: output},
	}
	var err error
	p.ID, err = m.ProcessStorage.CreateProcess(ctx, p)
	if err != nil {
		return nil, err
	}
	return m.loadProcess(ctx, p)
}

// GetEmitterEnvelopeStatus returns the status of the process triggered
// by the given emitter/envelope
func (m *Manager) GetEmitterEnvelopeStatus(ctx context.Context, emitter *api.Actor, envelopeID api.UUID) api.Process_Status {
	p, ok := m.ProcessStorage.GetProcessByTrigger(ctx, emitter, envelopeID)
	if !ok {
		return api.Process_NOSTATUS
	}
	return m.ProcessStorage.GetState(ctx, p.ID).Status
}

// Get returns a process from its ID
func (m *Manager) Get(ctx context.Context, id api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcess(ctx, id)
	if !ok {
		return nil, fmt.Errorf("No such process: %s", id)
	}
	return m.loadProcess(ctx, p)
}

// GetByTrigger returns the process matching the given trigger emitter/envelope
func (m *Manager) GetByTrigger(
	ctx context.Context, emitter *api.Actor, envelopeID api.UUID,
) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(ctx, emitter, envelopeID)
	if !ok {
		return nil, fmt.Errorf("No process for trigger (%s, %s)", emitter.ID, envelopeID)
	}
	return m.loadProcess(ctx, p)
}

// Replay creates a new process that replays the current process.
// The process must be terminated, as well as all the processes of the same
// group. Also, the trigger envelope must be complete and valid.
// If match is true, the MatchPipeline is called again. If false, the same
// pipeline is used, even if deactivated.
// match is ignored if the process never matched a pipeline, in which case
// MatchPipeline is always called.
func (m *Manager) Replay(ctx context.Context, process *Process, envelopeStorage storage.EnvelopeStorage, match bool) (*Process, error) {
	// First make sure the trigger envelope is complete
	status, err := envelopeStorage.ReceptionStatus(ctx, process.TriggerEnvelopeID)
	if err != nil {
		return nil, err
	}

	if status != storage.EnvelopeStorageReceptionComplete {
		return nil, fmt.Errorf("Trigger envelope is not complete (state is: %s)", status)
	}

	msgTypes, exists := envelopeStorage.GetEventTypes(ctx, process.TriggerEnvelopeID)
	if !exists {
		panic("Envelope must exist")
	}

	if process.Pipeline == nil {
		match = true
	}

	// TODO verify that no other process in the same group is running (or not ?
	// needs some thinking)

	groupID := process.GroupID
	if groupID.IsZero() {
		groupID = process.ID
	}
	np := storage.Process{
		GroupID:           groupID,
		ReplayOf:          process.ID,
		TriggerEmitterID:  process.TriggerEmitterID,
		TriggerEnvelopeID: process.TriggerEnvelopeID,
	}

	if !match {
		np.PipelineID = process.Pipeline.GetIDAsUUID()
		np.SourceOutput = process.SourceRef
	}

	id, err := m.ProcessStorage.CreateProcess(ctx, np)
	if err != nil {
		return nil, err
	}
	np.ID = id

	newProcess, err := m.loadProcess(ctx, np)
	if err != nil {
		// TODO set the process state to 'error' in the storage
		return nil, err
	}

	if match {
		if err := newProcess.MatchPipeline(ctx, process.SourceRef.Output, msgTypes); err != nil {
			return newProcess, err
		}
	}

	if _, err := newProcess.GetTargets(ctx, newProcess.SourceRef, newProcess.TriggerEnvelopeID, false); err != nil {
		return newProcess, err
	}
	if err := newProcess.UpdateReceiveStatus(
		ctx, process.TriggerEnvelopeID, process.TriggerEmitterID, newProcess.SourceRef, storage.ReceptionDone); err != nil {
		return newProcess, err
	}

	return newProcess, nil
}

// ListPostMortem returns dead processes
func (m *Manager) ListPostMortem(
	ctx context.Context, level api.LogLevel, status ...api.PMProcess_Status,
) ([]*Process, error) {
	var (
		postMortemList = m.ProcessStorage.QueryPostMortem(ctx, level, status...)
		result         = make([]*Process, len(postMortemList))
	)

	for i, sProcess := range postMortemList {
		process, err := m.loadProcess(ctx, sProcess)
		if err != nil {
			return nil, err
		}
		result[i] = process
	}

	return result, nil
}

// Purge removes any trace of the process in the process storage.
func (m *Manager) Purge(ctx context.Context, id api.UUID) error {
	// TODO drop the logs
	purged, err := m.ProcessStorage.PurgeProcess(ctx, []api.UUID{id})
	if err != nil {
		return err
	}
	if len(purged) != 1 {
		return fmt.Errorf("No terminated process with this id")
	}
	return nil
}

// PurgeMany removes any trace of the processes in the process storage.
func (m *Manager) PurgeMany(ctx context.Context, ids []api.UUID) ([]api.UUID, error) {
	return m.ProcessStorage.PurgeProcess(ctx, ids)
}
