// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package server

import (
	"context"

	"xbus.io/xbus/services"
	"xbus.io/xbus/services/broker"
	"xbus.io/xbus/services/clientapi"
	"xbus.io/xbus/services/control"
	"xbus.io/xbus/services/core"
	"xbus.io/xbus/services/director"
	"xbus.io/xbus/services/envelopestorage"
	"xbus.io/xbus/services/metrics"
	"xbus.io/xbus/services/postmortem"
	"xbus.io/xbus/services/processstate"
	"xbus.io/xbus/services/registration"
	"xbus.io/xbus/services/sentinel"
	"xbus.io/xbus/services/sentrycapture"
)

func (s *Server) servicesOptions() services.Options {
	return services.Options{
		Conn:   s.natsConn,
		Logger: s.log,
	}
}

func (s *Server) initServiceSentryCapture(ctx context.Context) *sentrycapture.Service {
	return sentrycapture.Options{
		Options: s.servicesOptions(),
		Hub:     s.sentryHub,
	}.NewService(ctx)
}

func (s *Server) initServiceBroker(ctx context.Context) *broker.Service {
	return broker.New(ctx, broker.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		EnvelopeStorage: s.envelopeStorage,
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
	})
}

func (s *Server) initServiceClientAPI(ctx context.Context) *clientapi.Service {
	return clientapi.NewService(ctx, clientapi.Options{
		Options:        s.servicesOptions(),
		AccountStorage: s.accountStorage,
		ActorStorage:   s.actorStorage,
		SessionStorage: s.sessionStorage,
	})
}

func (s *Server) initServiceControl(ctx context.Context) *control.Service {
	return control.NewService(ctx,
		control.Options{
			Options:         s.servicesOptions(),
			ServerCert:      s.clientCA,
			ServerKey:       s.clientCAKey,
			PipelineManager: s.pipelineManager,
			ProcessStorage:  s.processStorage,
			ProcessManager:  s.processManager,
			AccountStorage:  s.accountStorage,
			PasswordStorage: s.passwordStorage,
			ActorStorage:    s.actorStorage,
			EnvelopeStorage: s.envelopeStorage,
			LogStorage:      s.logStorage,
			JobStorage:      s.jobStorage,
		})
}

func (s *Server) initServiceCore(ctx context.Context) *core.Service {
	return core.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		ProcessStorage:  s.processStorage,
		ProcessManager:  s.processManager,
		PipelineManager: s.pipelineManager,
		LogStorage:      s.logStorage,
		EnvelopeStorage: s.envelopeStorage,
	}.NewService(ctx)
}

func (s *Server) initServiceDirector(ctx context.Context) *director.Service {
	return director.New(ctx, director.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		EnvelopeStorage: s.envelopeStorage,
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
		JobStorage:      s.jobStorage,
	})
}

func (s *Server) initServiceEnvelopeStorage(ctx context.Context) *envelopestorage.Service {
	opts := envelopestorage.Options{
		Options:         s.servicesOptions(),
		EnvelopeStorage: s.envelopeStorage,
		JobStorage:      s.jobStorage,
	}
	return opts.NewService(ctx)
}

func (s *Server) initServiceMetrics(ctx context.Context) (*metrics.Service, error) {
	opts := metrics.Options{
		Options: s.servicesOptions(),

		ServerCA: s.serverCert.Raw,
		ClientCA: s.clientCA.Raw,

		Hub: s.sentryHub,

		MetricsDelayHint: s.options.MetricsDelayHint,
	}
	svc, err := opts.NewService(ctx)
	if err != nil {
		return nil, err
	}
	s.metricsGatherer = svc.Gatherer()
	return svc, err
}

func (s *Server) initServiceProcessState(ctx context.Context) *processstate.Service {
	return processstate.Options{
		Options:         s.servicesOptions(),
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
		EnvelopeStorage: s.envelopeStorage,
		ActorStorage:    s.actorStorage,
	}.NewService(ctx)
}

func (s *Server) initServicePostmortem(ctx context.Context) *postmortem.Service {
	return postmortem.Options{
		Options:        s.servicesOptions(),
		LogStorage:     s.logStorage,
		ProcessManager: s.processManager,
	}.NewService(ctx)
}

func (s *Server) initServiceRegistration(ctx context.Context) *registration.Service {
	return registration.NewService(ctx,
		registration.Options{
			Options:        s.servicesOptions(),
			AccountStorage: s.accountStorage,
			ActorStorage:   s.actorStorage,
			ServerCA:       s.serverCert.Raw,
		})
}

func (s *Server) initServiceSentinel(ctx context.Context) *sentinel.Service {
	return sentinel.Options{
		Options:      s.servicesOptions(),
		Settings:     s.options.ServiceSentinel,
		ActorStorage: s.actorStorage,
	}.NewService(ctx)
}

func (s *Server) setupServices(ctx context.Context) error {
	s.ServiceManager = services.NewManager(s.log)

	if s.sentryHub != nil {
		s.ServiceManager.Add(s.initServiceSentryCapture(ctx))
	}

	s.ServiceManager.Add(s.initServiceCore(ctx))

	if s.options.ServeSentinel {
		s.ServiceManager.Add(s.initServiceSentinel(ctx))
	}

	if s.options.ServePostmortem {
		s.ServiceManager.Add(s.initServicePostmortem(ctx))
	}

	if s.options.ServeBroker {
		s.ServiceManager.Add(s.initServiceBroker(ctx))
	}

	if s.options.ServeClientAPI {
		s.ServiceManager.Add(s.initServiceClientAPI(ctx))
	}

	if s.options.ServeControl {
		s.ServiceManager.Add(s.initServiceControl(ctx))
	}

	if s.options.ServeDirector {
		s.ServiceManager.Add(s.initServiceDirector(ctx))
	}

	if s.options.ServeEnvelopeStorage {
		s.ServiceManager.Add(s.initServiceEnvelopeStorage(ctx))
	}

	if s.options.ServeProcessState {
		s.ServiceManager.Add(s.initServiceProcessState(ctx))
	}

	if s.options.ServeRegistration {
		s.ServiceManager.Add(s.initServiceRegistration(ctx))
	}

	if s.options.MetricsHTTPPort != 0 && s.options.ServeMetrics {
		svc, err := s.initServiceMetrics(ctx)
		if err != nil {
			return err
		}
		s.ServiceManager.Add(svc)
	}
	return nil
}
