// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package control

import (
	"context"
	"errors"

	"gopkg.in/yaml.v2"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/storage"
)

type pipelineServer struct {
	*Service
}

func (s pipelineServer) Query(
	ctx context.Context, req *control.PipelineQueryRequest,
) (*control.PipelineQueryReply, error) {
	graphs := s.o.PipelineManager.Query(ctx, req.Name, "", false)
	var resp control.PipelineQueryReply
	for _, graph := range graphs {
		resp.PipelineList = append(resp.PipelineList,
			&api.PipelineInfo{
				ID:      graph.ID,
				Name:    graph.Name,
				Version: graph.Version,
				Status:  graph.Status,
			})
	}
	return &resp, nil
}

func (s pipelineServer) Save(
	ctx context.Context, req *control.PipelineSaveRequest,
) (
	*control.PipelineSaveReply, error,
) {
	pl := pipeline.NewPipeline(storage.NewPipeline())
	if err := yaml.Unmarshal([]byte(req.Graph), &pl.Graph); err != nil {
		err = api.ClientProtoError("Error decoding graph: %s", err)
		return nil, err
	}

	if req.Info.GetIDAsUUID().IsZero() {
		pipes := s.o.PipelineManager.Query(
			ctx, req.Info.Name, req.Info.Version, false)
		if len(pipes) > 1 {
			err := api.ClientProtoError(
				"Multiple pipeline matches: %s-%s", req.Info.Name, req.Info.Version)
			return nil, err
		}
		if len(pipes) == 1 {
			pl.ID = pipes[0].ID
		}
		pl.Name = req.Info.Name
		pl.Version = req.Info.Version
	} else {
		curPipeline, exists := s.o.PipelineManager.GetPipeline(
			ctx, req.Info.GetIDAsUUID())
		if exists {
			err := api.ClientProtoError(
				"No such pipeline: %s", req.Info.ID)
			return nil, err
		}
		pl.ID = req.Info.ID
		pl.Name = curPipeline.Name
		pl.Version = curPipeline.Version
	}

	var (
		resp = &control.PipelineSaveReply{}
		err  error
	)
	validator := pipeline.NewValidator(pl.Graph)

	for _, e := range validator.Validate() {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	for _, e := range validator.ValidateActors(ctx, s.o.ActorStorage) {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	pl, err = s.o.PipelineManager.Save(ctx, pl)
	if err != nil {
		err = api.ClientProtoError("Save pipeline error: %s", err)
		return nil, err
	}

	var b []byte
	b, err = yaml.Marshal(pl.Graph)
	if err != nil {
		err = api.ServerProtoError(err)
		return nil, err
	}
	resp.IsValid = len(resp.ValidationMessages) == 0
	resp.Graph = string(b)
	resp.Info = pl.PipelineInfo
	return resp, nil
}

func (s pipelineServer) SetStatus(
	ctx context.Context, req *api.PipelineInfo,
) (*control.PipelineSetStatusReply, error) {
	info := api.PipelineInfo{ID: req.ID}
	var pl *pipeline.Pipeline
	if info.GetIDAsUUID().IsZero() {
		pipes := s.o.PipelineManager.Query(ctx, req.Name, req.Version, false)
		if len(pipes) == 0 {
			return nil, api.ClientProtoError(
				"No such graph: %s-%s", req.Name, req.Version)
		}
		if len(pipes) > 1 {
			return nil, api.ClientProtoError("Multiple graphs found")
		}
		pl = pipes[0]
		info.ID = pl.ID
		info.Name = pl.Name
		info.Version = pl.Version
		info.Status = pl.Status
	} else {
		var ok bool
		pl, ok = s.o.PipelineManager.GetPipeline(ctx, info.GetIDAsUUID())
		if !ok {
			return nil, api.ClientProtoError(
				"No such graph: %s", info.ID)
		}
		info.ID = pl.ID
		info.Name = pl.Name
		info.Version = pl.Version
		info.Status = pl.Status
	}
	resp := &control.PipelineSetStatusReply{
		Info: &info,
	}
	if setStatusErr := s.o.PipelineManager.SetStatus(ctx, pl, req.Status); setStatusErr != nil {
		var list pipeline.ValidationErrorList
		if errors.As(setStatusErr, &list) {
			resp.IsValid = false
			for _, e := range list {
				resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
			}
		} else {
			return nil, api.ClientProtoError("%s", setStatusErr)
		}
	} else {
		resp.IsValid = true
		resp.Info.Status = req.Status
	}
	return resp, nil
}

func (s pipelineServer) GetGraph(
	ctx context.Context, req *api.PipelineInfo,
) (
	*api.StringReply, error,
) {
	pipeID := req.GetIDAsUUID()
	var result *pipeline.Pipeline
	if pipeID.IsZero() {
		if req.GetName() == "" {
			return nil, api.ClientProtoError("Invalid graph query: must have an ID or a Name")
		}
		pipes := s.o.PipelineManager.Query(
			ctx,
			req.GetName(), req.GetVersion(),
			req.GetStatus() == api.PipelineInfo_ACTIVE)
		switch len(pipes) {
		case 0:
			return nil, api.ClientProtoError("No such graph")
		case 1:
			pl := pipes[0]
			result = pl
		default:
			var active *pipeline.Pipeline
			// get the draft version if unique
			for i := range pipes {
				pl := pipes[i]
				switch pl.Status {
				case api.PipelineInfo_DRAFT:
					if result != nil {
						return nil, api.ClientProtoError(
							"Graph %s has multiple drafts: please specify the version",
							req.GetName())
					}
					result = pl
				case api.PipelineInfo_ACTIVE:
					active = pl
				}
			}
			if result == nil && active != nil {
				result = active
			}
			if result == nil {
				return nil, api.ClientProtoError(
					"Graph %s has no draft of active version: please specify the version you want",
					req.GetName(),
				)
			}
		}
	} else {
		pl, exists := s.o.PipelineManager.GetPipeline(ctx, req.GetIDAsUUID())
		if !exists {
			return nil, api.ClientProtoError("No such graph")
		}
		result = pl
	}
	b, err := yaml.Marshal(result.Graph)
	if err != nil {
		panic(err)
	}
	return &api.StringReply{
		Value: string(b),
	}, nil
}
