// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package postmortem

import (
	"context"
	"time"

	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "postmortem"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Options is used to create a 'processstate' service
type Options struct {
	services.Options

	LogStorage     storage.LogStorage
	ProcessManager *process.Manager
}

// NewService creates a Service
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o Options
}

// Startup starts the service
// TODO check if any process is done with warning/error and has a default postmortem, and
// initialize them (the process most likely stopped when the service was off)
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	coreEvents := api.NewCoreEventsClient(s.Conn)
	// TODO do a queue subscription (needs so rework of the 'norequest' branch in nrpc
	sub, err := coreEvents.ProcessSubscribe(s.handleCoreEventProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)
	return nil
}

func (s *Service) handleCoreEventProcess(evt *api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	process, err := s.o.ProcessManager.Get(s.Ctx, evt.Process.GetIDAsUUID())
	if err != nil {
		panic(err)
	}

	if evt.Process.Status == api.Process_ERROR {
		if err := process.SetPostMortemState(s.Ctx, storage.PostMortemState{
			Level:   api.LogLevel_ERROR,
			Status:  api.PMProcess_NEW,
			Comment: "",
			Changed: time.Now(),
		}); err != nil {
			s.o.Options.Logger.Err(err).Msg("Error setting post mortem state")
		}
	} else if evt.Process.Status == api.Process_DONE {
		logs := s.o.LogStorage.GetProcessMessages(s.Ctx, process.ID, api.LogLevel_WARNING)
		if len(logs) != 0 {
			if err := process.SetPostMortemState(s.Ctx, storage.PostMortemState{
				Level:   api.LogLevel_WARNING,
				Status:  api.PMProcess_NEW,
				Comment: "",
				Changed: time.Now(),
			}); err != nil {
				s.o.Options.Logger.Err(err).Msg("Error setting post mortem state")
			}
		} else {
			if err := process.SetPostMortemState(s.Ctx, storage.PostMortemState{
				Level:   api.LogLevel_NOTICE,
				Status:  api.PMProcess_DEFAULT,
				Comment: "",
				Changed: time.Now(),
			}); err != nil {
				s.o.Options.Logger.Err(err).Msg("Error setting post mortem state")
			}
		}
	}
}
