// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package processstate

import (
	"context"
	"fmt"

	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "processstate"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Options is used to create a 'processstate' service
type Options struct {
	services.Options

	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager
	EnvelopeStorage storage.EnvelopeStorage
	ActorStorage    storage.ActorStorage
}

// NewService creates a Service
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o Options

	api          *api.Client
	processState *api.ProcessStateHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	s.processState = api.NewProcessStateHandler(
		s.Ctx, s.Conn, s,
	)
	s.processState.SetEncodings([]string{"protobuf", "json"})

	s.api = api.NewClient(s.Conn)
	sub, err := s.api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	return s.SubscribeAll(
		s.HandlerQueueSubscribe(s.processState, QueueName),
	)
}

func (s *Service) handleCoreEventsProcess(evt *api.CoreEvent) {
	s.TaskAdd()
	defer s.TaskDone()

	envelopeState := api.EmitterEnvelopeState{
		ProcessID: evt.Process.ID,
		Status:    evt.Process.Status,
	}

	for _, entry := range s.o.LogStorage.GetProcessMessages(
		s.Ctx, evt.Process.GetIDAsUUID(), api.LogLevel_ERROR,
	) {
		envelopeState.Errors = append(envelopeState.Errors, entry.Message)
	}

	// load the trigger envelope head
	envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
		s.Ctx, evt.Process.GetTriggerEnvelopeIDAsUUID(), storage.EnvelopeStartPos(), 100*1024,
	)
	if err != nil {
		s.Log.Err(err).
			Str("process-id", evt.Process.GetIDAsUUID().String()).
			Str("envelope-id", evt.Process.GetTriggerEnvelopeIDAsUUID().String()).
			Msg("Could not read process trigger envelope")
	} else {
		envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
			Fragment: envelope,
			Position: storage.EncodePosition(evt.Process.GetTriggerEnvelopeIDAsUUID(), position),
		}
	}

	if evt.Process.Status == api.Process_DONE {
		// Check for a response
		p, err := s.o.ProcessManager.Get(s.Ctx, evt.Process.GetIDAsUUID())
		if err != nil {
			panic(err)
		}
		responseID := p.GetResponse(s.Ctx)
		if !responseID.IsZero() {
			envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
				s.Ctx, responseID, storage.EnvelopeStartPos(), 512*1024,
			)
			if err != nil {
				s.Log.Err(err).
					Str("process-id", evt.Process.GetIDAsUUID().String()).
					Str("envelope-id", responseID.String()).
					Msg("Could not read process response envelope")
			}
			envelopeState.Response = &api.EnvelopeFragmentAndPosition{
				Fragment: envelope,
				Position: storage.EncodePosition(responseID, position),
			}
		}
	}

	if err := s.processState.EnvelopeStatesPublish(
		evt.Process.GetTriggerEmitterIDAsUUID().String(),
		evt.Process.GetTriggerEnvelopeIDAsUUID().String(),
		&envelopeState,
	); err != nil {
		s.Log.Err(err).Msg("Error sending process state")
	}
}

// GetEnvelopeState implements ProcessState.GetEnvelopeState
func (s *Service) GetEnvelopeState(
	ctx context.Context, emitterID string, req *api.GetEnvelopeStateRequest,
) (
	*api.EmitterEnvelopeState, error,
) {
	emitterUUID, err := api.UUIDFromString(emitterID)
	if err != nil {
		return nil, fmt.Errorf("invalid emitterID: %w", err)
	}
	emitter, ok := s.o.ActorStorage.Get(ctx, emitterUUID)
	if !ok {
		return nil, fmt.Errorf("unknown emitter: %w", err)
	}
	process, err := s.o.ProcessManager.GetByTrigger(ctx, emitter, req.GetIDAsUUID())
	if err != nil {
		return nil, err
	}

	envelopeState := api.EmitterEnvelopeState{
		ProcessID:        process.ID.ToBytes(),
		ProcessCreatedAt: timestamppb.New(process.CreatedAt),
		Status:           process.GetStatus(ctx),
	}

	// load the trigger envelope head
	envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
		ctx, process.TriggerEnvelopeID, storage.EnvelopeStartPos(), 100*1024,
	)
	if err != nil {
		s.Log.Err(err).
			Str("process-id", process.ID.String()).
			Str("envelope-id", process.TriggerEnvelopeID.String()).
			Msg("Could not read process trigger envelope")
	} else {
		envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
			Fragment: envelope,
			Position: storage.EncodePosition(process.TriggerEnvelopeID, position),
		}
	}

	if envelopeState.Status == api.Process_DONE {
		// Check for a response
		responseID := process.GetResponse(ctx)
		if !responseID.IsZero() {
			envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
				ctx, responseID, storage.EnvelopeStartPos(), 100*1024,
			)
			if err != nil {
				s.Log.Err(err).
					Str("process-id", process.ID.String()).
					Str("envelope-id", responseID.String()).
					Msg("Could not read process response envelope")
			}
			envelopeState.Response = &api.EnvelopeFragmentAndPosition{
				Fragment: envelope,
				Position: storage.EncodePosition(responseID, position),
			}
		}
	}

	if envelopeState.Status == api.Process_ERROR {
		for _, entry := range s.o.LogStorage.GetProcessMessages(
			ctx, envelopeState.GetProcessIDAsUUID(), api.LogLevel_ERROR,
		) {
			envelopeState.Errors = append(envelopeState.Errors, entry.Message)
		}
	}

	return &envelopeState, nil
}

// AckResult mark a process result as acked
func (s *Service) AckResult(
	ctx context.Context, emitterID string, req *api.AckResultRequest,
) error {
	emitterUUID, err := api.UUIDFromString(emitterID)
	if err != nil {
		return fmt.Errorf("invalid emitterID: %w", err)
	}
	_, ok := s.o.ActorStorage.Get(ctx, emitterUUID)
	if !ok {
		return fmt.Errorf("unknown emitter: %w", err)
	}
	process, err := s.o.ProcessManager.Get(ctx, req.GetProcessIDAsUUID())
	if err != nil {
		return err
	}
	if process.TriggerEmitterID != emitterUUID {
		return fmt.Errorf("This emitter has no such process ID")
	}
	s.o.ProcessManager.ProcessStorage.AckResult(ctx, req.GetProcessIDAsUUID())

	return nil
}

// List the process of this emitter
func (s *Service) List(
	ctx context.Context, emitterID string, req *api.ProcessFilter,
	send func(*api.EmitterEnvelopeStateList),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	req.TriggerEmitterID = api.UUIDSlice{
		api.MustUUIDFromString(emitterID),
	}.ToBytes()

	pList := s.o.ProcessManager.ProcessStorage.Query(ctx, req)
	select {
	case <-ctx.Done():
		return nil
	default:
	}
	batch := make([]*api.EmitterEnvelopeState, 0, 200)
	for _, p := range pList {
		state := s.o.ProcessManager.ProcessStorage.GetState(ctx, p.ID)
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		envelopeState := api.EmitterEnvelopeState{
			ProcessID: p.ID.ToBytes(),
			Status:    state.Status,
		}
		envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
			ctx, p.TriggerEnvelopeID, storage.EnvelopeStartPos(), 512*1024,
		)
		if err != nil {
			s.Log.Err(err).
				Str("process-id", p.ID.String()).
				Str("envelope-id", p.TriggerEnvelopeID.String()).
				Msg("Could not read process trigger envelope")
		} else {
			envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
				Fragment: envelope,
				Position: storage.EncodePosition(p.TriggerEnvelopeID, position),
			}
		}
		if envelopeState.Status == api.Process_DONE {
			// Check for a response
			responseID := state.ResponseEnvelopeID
			if !responseID.IsZero() {
				envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
					ctx, responseID, storage.EnvelopeStartPos(), 512*1024,
				)
				if err != nil {
					s.Log.Err(err).
						Str("process-id", p.ID.String()).
						Str("envelope-id", responseID.String()).
						Msg("Could not read process response envelope")
				}
				envelopeState.Response = &api.EnvelopeFragmentAndPosition{
					Fragment: envelope,
					Position: storage.EncodePosition(responseID, position),
				}
			}
		}
		if envelopeState.Status == api.Process_ERROR {
			for _, entry := range s.o.LogStorage.GetProcessMessages(
				ctx, envelopeState.GetProcessIDAsUUID(), api.LogLevel_ERROR,
			) {
				envelopeState.Errors = append(envelopeState.Errors, entry.Message)
			}
		}
		batch = append(batch, &envelopeState)
		if len(batch) == 200 {
			send(&api.EmitterEnvelopeStateList{List: batch})
			batch = batch[0:0]
		}
	}
	if len(batch) != 0 {
		send(&api.EmitterEnvelopeStateList{List: batch})
	}
	return nil
}
