// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

// Package sentrycapture provides the 'core' service, which must be started once and only once
// in every server process
package sentrycapture

import (
	"context"
	"encoding/json"
	"fmt"
	"strconv"
	"time"

	sentry "github.com/getsentry/sentry-go"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/services"
)

func makeJSONReady(value interface{}) interface{} {
	switch value := value.(type) {
	case map[interface{}]interface{}:
		res := make(map[string]interface{})

		for key, item := range value {
			item = makeJSONReady(item)
			switch k := key.(type) {
			case string:
				res[k] = item
			default:
				res[fmt.Sprintf("%v", k)] = item
			}
		}

		return res
	case []interface{}:
		var res []interface{}
		for _, item := range value {
			item = makeJSONReady(item)
			res = append(res, item)
		}
		return res
	default:
		return value
	}
}

func extractVar(value interface{}, path ...string) interface{} {
	if len(path) == 0 {
		return value
	}
	key := path[0]
	subPath := path[1:]
	switch value := value.(type) {
	case map[string]interface{}:
		return extractVar(value[key], subPath...)
	case map[interface{}]interface{}:
		return extractVar(value[key], subPath...)
	case []interface{}:
		iKey, err := strconv.Atoi(key)
		if err != nil {
			return nil
		}
		return value[iKey]
	default:
		return nil
	}
}

func extractVarString(value interface{}, path ...string) string {
	s, _ := extractVar(value, path...).(string)
	return s
}

// Options is used to create a 'sentrycapture' service
type Options struct {
	services.Options

	Hub *sentry.Hub
}

// NewService creates a Emitter
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, "sentrycapture", o.Options),
		o:           o,
		actorNames:  make(map[api.UUID]string),
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options

	actorNames map[api.UUID]string
}

func (s *Service) refreshActorNames() {
	// We may prefer to use a ActorMap, but I have a feeling its overkill
	// as the actor names are not likely to change
	client := api.NewClient(s.Conn)
	ctx, cancel := context.WithTimeout(s.Ctx, 5*time.Second)
	defer cancel()
	if err := client.Sentinel.ListActors(ctx,
		func(ctx context.Context, reply *api.ClientGetActorsReply) {
			for _, actor := range reply.Actors {
				s.actorNames[actor.GetIDAsUUID()] = actor.Name
			}
		},
	); err != nil {
		s.Log.Err(err).Msg("failed to refresh the actor names list")
	}
}

func (s *Service) getActorName(id api.UUID) string {
	if name, ok := s.actorNames[id]; ok {
		return name
	}
	s.refreshActorNames()
	if name, ok := s.actorNames[id]; ok {
		return name
	}
	return id.String()
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	api := api.NewClient(s.Conn)
	sub, err := api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)
	return nil
}

func (s *Service) handleCoreEventsProcess(evt *api.CoreEvent) {
	if evt.Process.Status == api.Process_ERROR {
		c := control.NewProcessClient(s.Conn)

		var buf []byte

		if err := c.ExportStream(
			s.Ctx,
			&control.ProcessExportRequest{
				ProcessIDs:        api.UUIDSlice{evt.Process.GetIDAsUUID()}.ToBytes(),
				ExportEnvelope:    true,
				ExportJobs:        false,
				ExportLogs:        true,
				ExportProcessLogs: true,
				EnvelopeSizeLimit: 1024 * 10,
				LogLevel:          api.LogLevel_NOTICE,
			},
			func(ctx context.Context, res *control.OctetStreamList) {
				for _, chunk := range res.GetChunks() {
					buf = append(buf, chunk.Data...)
				}
			},
		); err != nil {
			s.Log.Err(err).
				Str("process-id", evt.Process.GetIDAsUUID().String()).
				Msg("could not export process")
			return
		}

		var data []map[string]interface{}

		if err := json.Unmarshal(buf, &data); err != nil {
			s.Log.Err(err).Msg("process export could not be parsed as json")
			return
		}

		if len(data) != 1 {
			s.Log.Error().Msgf("Expected 1 process in the export, got %d", len(data))
			return
		}

		type processlogsLogMessage struct {
			Level api.LogLevel
			Text  string
		}
		type processlogsLog struct {
			ActorID api.UUID
			Message processlogsLogMessage
		}
		type processlogs struct {
			Logs []processlogsLog
		}
		var logs []processlogs

		if err := json.Unmarshal(buf, &logs); err != nil {
			s.Log.Warn().Err(err).Msg("Could not read logs in the export")
			return
		}

		var (
			tags   = make(map[string]string)
			extra  = make(map[string]interface{})
			reason = "N/A"
		)

		tags["xbus.emitter"] = extractVarString(data[0], "trigger", "emitter", "name")

		if len(logs) == 1 {
			for i := range logs[0].Logs {
				log := &logs[0].Logs[i]
				if log.Message.Level == api.LogLevel_ERROR {
					tags["xbus.actor"] = s.getActorName(log.ActorID)
					reason = log.Message.Text
					break
				}
			}
		}
		if reason == "N/A" {
			if value := extractVarString(data[0], "state", "status_reason"); value != "" {
				reason = value
			} else {
				s.Log.Error().Msg("Could not extract StatusReason")
			}
		}

		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "name"); value != "" {
			tags["xbus.pipeline.name"] = value
		}
		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "version"); value != "" {
			tags["xbus.pipeline.version"] = value
		}

		for name, data := range data[0] {
			data = makeJSONReady(data)
			if name == "id" {
				name = "process_id"
			}
			name = "xbus." + name
			extra[name] = data
		}

		text := "[process]"
		if tags["xbus.pipeline.name"] != "" {
			text += fmt.Sprintf("[%s-%s]",
				tags["xbus.pipeline.name"], tags["xbus.pipeline.version"],
			)
		} else {
			text += fmt.Sprintf("[from %s]", tags["xbus.emitter"])
		}
		text += " " + reason

		event := sentry.NewEvent()
		event.Message = text
		event.Tags = tags
		event.Extra = extra
		event.Level = sentry.LevelError

		s.o.Hub.CaptureEvent(event)
	}
}
