// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package services

import (
	"context"
	"fmt"
	"runtime/debug"
	"sync"
	"time"

	"github.com/nats-io/nats.go"
	"github.com/rs/zerolog"
)

// Options is the common options of xbus services
type Options struct {
	Conn   *nats.Conn
	Logger zerolog.Logger
}

// Service is the common interface of all the services
type Service interface {
	// GetName returns the service name
	GetName() string

	// Startup will do what is necessary for the service to be up and running
	Startup() error

	// Shutdown should stop the service and release any resource allocated by
	// Startup
	Shutdown() error
}

// BaseService is a base struct for implementing services.
type BaseService struct {
	Name    string
	Ctx     context.Context
	rootCtx context.Context
	stop    func()

	Log zerolog.Logger

	Conn *nats.Conn
	subs []*nats.Subscription

	Tasks sync.WaitGroup
}

// NewBaseService initialize a BaseService
func NewBaseService(ctx context.Context, name string, o Options) BaseService {
	return BaseService{
		Name:    name,
		rootCtx: ctx,
		Log:     o.Logger.With().Str("service", name).Logger(),
		Conn:    o.Conn,
	}
}

// ping
// subscribe to xbus.ping.services.<name>

// isAlone
// send a ping and check how many instances reply.
// if only one (us), return true

// GetName returns the service nams
func (s *BaseService) GetName() string {
	return s.Name
}

func (s *BaseService) Startup() error {
	s.Ctx, s.stop = context.WithCancel(context.Background())
	go func() {
		select {
		case <-s.rootCtx.Done():
			if err := s.Shutdown(); err != nil {
				s.Log.Err(err).Msg("emergency shutdown failed")
			}
		case <-s.Ctx.Done():
		}
	}()
	return nil
}

// Shutdown is a default implementation that calls UnsubscribeAll()
func (s *BaseService) Shutdown() error {
	err := s.UnsubscribeAll()
	timer := time.AfterFunc(time.Second, s.stop)
	s.Tasks.Wait()
	defer func() {
		timer.Stop()
		s.stop()
	}()
	return err
}

func (s *BaseService) BlindShutdown() {
	if err := s.Shutdown(); err != nil {
		s.Log.Err(err).Msg("shutdown failed")
	}
}

// AddSubscription save a subscription to be unsubscribed with UnsubscribeAll
func (s *BaseService) AddSubscription(sub *nats.Subscription) {
	s.subs = append(s.subs, sub)
}

// SubscriptionDef is a definition of a nats subscription
type SubscriptionDef struct {
	Subject string
	Queue   string
	Handler nats.MsgHandler
}

// Subscribe returns a SubscriptionDef for simple subscription
func Subscribe(subject string, handler nats.MsgHandler) SubscriptionDef {
	return SubscriptionDef{
		Subject: subject,
		Handler: handler,
	}
}

// QueueSubscribe returns a SubscriptionDef for queue subscription
func QueueSubscribe(subject, queue string, handler nats.MsgHandler) SubscriptionDef {
	return SubscriptionDef{
		Subject: subject,
		Queue:   queue,
		Handler: handler,
	}
}

// NRPCHandler ...
type NRPCHandler interface {
	Subject() string
	Handler(*nats.Msg)
}

// HandlerQueueSubscribe returnes a SubscriptionDef for a nrpc queue subscription
func (s *BaseService) HandlerQueueSubscribe(handler NRPCHandler, queue string) SubscriptionDef {
	return SubscriptionDef{
		Subject: handler.Subject(),
		Queue:   queue,
		Handler: func(msg *nats.Msg) {
			s.TaskAdd()
			defer s.TaskDone()
			handler.Handler(msg)
		},
	}
}

// SubscribeAll do multiple subscriptions at once and returns
// immediately on the first failing subscription.
func (s *BaseService) SubscribeAll(subdefs ...SubscriptionDef) error {
	for _, subdef := range subdefs {
		sub, err := s.Conn.QueueSubscribe(subdef.Subject, subdef.Queue, subdef.Handler)
		if err != nil {
			return err
		}
		s.AddSubscription(sub)
	}
	return nil
}

// UnsubscribeAll unsubscribe and forget all the subscriptions
func (s *BaseService) UnsubscribeAll() error {
	var lastError error
	for _, sub := range s.subs {
		err := sub.Unsubscribe()
		if err != nil {
			s.Log.Err(err).Str("subject", sub.Subject).Msg("error unsubscribing from")
			lastError = err
		}
	}
	s.subs = nil
	return lastError
}

// TaskAdd signal that a task just started.
// It will block service shutdown until TaskDone is called.
func (s *BaseService) TaskAdd() {
	s.Tasks.Add(1)
}

// TaskDone signal the end of a task, allowing the service to shutdown
// if no other task is running.
func (s *BaseService) TaskDone() {
	s.Tasks.Done()
}

// SafeTaskDone is a version of TaskDone that recover panics and log them.
func (s *BaseService) SafeTaskDone() {
	if r := recover(); r != nil {
		s.Log.Error().
			Str("stack", string(debug.Stack())).
			Str("error", fmt.Sprintf("%v", r)).
			Msg("caught panic in Tasks.Done()")
	}
	s.Tasks.Done()
}
