// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"context"
	"time"

	"xbus.io/go-xbus/v4/api"
)

// ProcessEventStatusHandler is the type of function that handle
// status changes
type ProcessEventStatusHandler func(p Process, old api.Process_Status, new api.Process_Status)

// NoOpProcessEventStatusHandler is a noop handler
func NoOpProcessEventStatusHandler(Process, api.Process_Status, api.Process_Status) {}

// ProcessLogEntryHandler is the type of functions that handle process log entries
type ProcessLogEntryHandler func(entries []ProcessLogEntry)

// VoidProcessLogEntryHandler is a noop ProcessLogEntryHandler
func VoidProcessLogEntryHandler([]ProcessLogEntry) {}

// Process contains the data necessary for process persistence
type Process struct {
	ID                api.UUID
	GroupID           api.UUID
	TriggerEmitterID  api.UUID
	TriggerEnvelopeID api.UUID
	CreatedAt         time.Time
	ReplayOf          api.UUID
	PipelineID        api.UUID
	SourceOutput      NodeOutputRef
}

// StateUpdater provides exclusive access to a process state
type StateUpdater interface {
	Get(context.Context) (ProcessState, error)
	Set(context.Context, ProcessState) error
	MarkLogEntry(ctx context.Context, id string, asError bool) error
	Unlock() error
}

// Emission identifies the emission of an envelope to a target
type Emission struct {
	ProcessID      api.UUID
	EnvelopeID     api.UUID
	TargetActor    api.UUID
	TargetInputRef NodeInputRef
}

// Reception identifies a reception of an envelop from an actor
type Reception struct {
	ProcessID       api.UUID
	EnvelopeID      api.UUID
	SourceActor     api.UUID
	SourceOutputRef NodeOutputRef
}

// ProcessNodeInput describes which envelope does on a given input
type ProcessNodeInput struct {
	Name       string
	EnvelopeID api.UUID
}

// ReceptionStatus describes the status of an envelope reception in
// the context of a process
type ReceptionStatus int

const (
	// ReceptionPending is when envelop reception is not yet started
	ReceptionPending ReceptionStatus = iota
	// ReceptionOngoing is when part(s) of the envelope were received, but not all
	ReceptionOngoing
	// ReceptionDone is when all the complete envelop has been receiced
	ReceptionDone
	// ReceptionError is when a reception error occurred
	ReceptionError
	// ReceptionClosed is when no envelope is to be expected
	ReceptionClosed
)

// ProcessNodeEnd signals the end of processing by a process node
type ProcessNodeEnd struct {
	Node   ProcessNode
	Status api.ActorProcessingState_Status
}

// ProcessLogEntry is the unique source of information for process tracking.
type ProcessLogEntry struct {
	ID               string
	ProcessID        api.UUID
	Timestamp        time.Time
	Emission         *Emission        `json:",omitempty"`
	EmissionPosition *EnvelopeReadPos `json:",omitempty"`
	Reception        *Reception       `json:",omitempty"`
	ReceptionStatus  ReceptionStatus  `json:",omitempty"`
	ProcessNodeEnd   *ProcessNodeEnd  `json:",omitempty"`
	ResultAck        bool             `json:",omitempty"`
}

// A ProcessNode is a live pipeline graph node. One graph node can correspond
// to several ProcessNode (especially consumer nodes)
type ProcessNode struct {
	ActorID api.UUID
	NodeID  string
}

//go:generate enum-marshaler -type ProcessNodeStatus -prefix ProcessNode -lower -tests

// ProcessNodeStatus is the process node status
type ProcessNodeStatus int

const (
	// ProcessNodeStatusPending is before node activation
	ProcessNodeStatusPending ProcessNodeStatus = iota
	// ProcessNodeActive is when any input or output is active
	ProcessNodeActive
	// ProcessNodeDone is when all input/output are done transmitting
	ProcessNodeDone
	// ProcessNodeUnreachable is when every path leading to the node
	// are closed and the node will never be active
	ProcessNodeUnreachable
	// ProcessNodeError is when an error occurred on the node
	ProcessNodeError
)

//go:generate enum-marshaler -type ProcessNodeIOStatus -prefix ProcessNodeIO -lower -tests

// ProcessNodeIOStatus reflects the status of the inputs or outputs of a node
type ProcessNodeIOStatus int

const (
	// ProcessNodeIOPending is when no io started yet
	ProcessNodeIOPending ProcessNodeIOStatus = iota
	// ProcessNodeIOStarted is when one or more io started emitting
	ProcessNodeIOStarted
	// ProcessNodeIOAllStarted is when all io started emitting
	ProcessNodeIOAllStarted
	// ProcessNodeIOClosed is when all the io is closed without any transmission
	ProcessNodeIOClosed
	// ProcessNodeIODone is when all io finished emitting
	ProcessNodeIODone
	// ProcessNodeIOError is when an error occurred on one of the io
	ProcessNodeIOError
)

// ProcessNodeIOStatusMap is a map of emission or reception statuses
type ProcessNodeIOStatusMap map[string]ProcessNodeIOStatus

// ProcessNodeState reflect the current state of a process node.
type ProcessNodeState struct {
	Status         ProcessNodeStatus
	Error          string
	Input          ProcessNodeIOStatus
	InputStatuses  ProcessNodeIOStatusMap
	Output         ProcessNodeIOStatus
	OutputStatuses ProcessNodeIOStatusMap
}

// ProcessNodeStateMap contains states of process nodes
type ProcessNodeStateMap map[ProcessNode]ProcessNodeState

// ProcessState reflect the current state of a process execution.
// It is completely updated by applying incoming logentries
type ProcessState struct {
	Status            api.Process_Status
	StatusChanged     time.Time
	StatusReason      string
	NodeStates        ProcessNodeStateMap
	NodeStatesChanged time.Time

	ResponseEnvelopeID api.UUID
	ResultAcked        bool
}

// PostMortemState describes the post-mortem state of a Process
type PostMortemState struct {
	Level   api.LogLevel
	Status  api.PMProcess_Status
	Comment string
	Changed time.Time
}

// ProcessSummary is a set of statistics about the process
type ProcessSummary []ProcessSummaryEntry

// ProcessSummaryEntry ...
type ProcessSummaryEntry struct {
	PipelineID  api.UUID
	EmitterID   api.UUID
	Status      api.Process_Status
	ResultAcked bool

	Count uint64
}

// ProcessStorage persists process data
type ProcessStorage interface {
	GetProcess(context.Context, api.UUID) (Process, bool)
	GetProcessByTrigger(context.Context, *api.Actor, api.UUID) (Process, bool)

	Query(context.Context, *api.ProcessFilter) []Process
	QuerySummary(context.Context, *api.ProcessFilter) (ProcessSummary, error)

	FindProcessForEnvelope(context.Context, []api.UUID) map[api.UUID]api.UUID

	CreateProcess(context.Context, Process) (api.UUID, error)
	SetPipeline(ctx context.Context, processID api.UUID, pipelineID api.UUID, sourceOutput NodeOutputRef)

	// GetTargets retrieve targets for an envelope from a routing
	// table. If the table has no matching entry, a callback is used to
	// determine the targets and if the envelope is the process response,
	// and the table is updated.

	// When the table is updated, ProcessLogEntry of type 'emission' are
	// generated

	// In some cases we do not want the table to be updated (more especially we do
	// not want log entries to be generated), so the function has a
	// 'noRouteTableUpdate' argument.

	GetTargets(
		ctx context.Context,
		processID api.UUID,
		source NodeOutputRef,
		envelopeID api.UUID,
		noRouteTableUpdate bool,
		calcTargets func() ([]*api.EnvelopeTarget, bool, error),
	) ([]*api.EnvelopeTarget, error)

	GetInputs(
		ctx context.Context,
		processID api.UUID,
		targetNode ProcessNode,
	) []ProcessNodeInput

	UpdateReceiveStatus(context.Context, Reception, ReceptionStatus) error

	ProcessNodeEnd(
		ctx context.Context,
		process api.UUID, node ProcessNode,
		status api.ActorProcessingState_Status) error

	GetState(context.Context, api.UUID) ProcessState
	GetStateUpdater(context.Context, api.UUID) StateUpdater
	SetStatus(context.Context, api.UUID, api.Process_Status, string) error

	// Generate a AckResult process log entry
	AckResult(context.Context, api.UUID)

	SetEventHandlers(ProcessEventStatusHandler)
	SetProcessLogEntryHandler(ProcessLogEntryHandler)

	GetUnprocessedProcessLogEntries(context.Context) []ProcessLogEntry
	GetProcessLogEntries(context.Context, api.UUID) []ProcessLogEntry

	QueryPostMortem(ctx context.Context, level api.LogLevel, status ...api.PMProcess_Status) []Process
	GetPostMortemState(context.Context, api.UUID) PostMortemState
	SetPostMortemState(context.Context, api.UUID, PostMortemState)

	// PurgeProcess removes all data of some processes.
	// The process must be terminated
	// Returns the list of process purged. If some ids are missing, the process
	// were non-existent or they had a non terminal status
	PurgeProcess(context.Context, []api.UUID) ([]api.UUID, error)
}
