// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"crypto/sha256"
	"database/sql"
	"encoding/hex"
	"errors"
	"fmt"
	"time"

	jsoniter "github.com/json-iterator/go"
	"github.com/lib/pq"
	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"google.golang.org/protobuf/proto"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
	"xbus.io/xbus/utils"
)

// NewEnvelopeStorageNG initialize a EnvelopeStorageNG
func NewEnvelopeStorageNG(model *model.Model, db *yago.DB, logger zerolog.Logger) *EnvelopeStorageNG {
	return &EnvelopeStorageNG{
		model:        model,
		db:           db,
		logger:       logger.With().Str("storage", "storage/sql/envelope-ng").Logger(),
		eventHandler: nil,
		tableSize:    NewTableSize("envelope_storage"),
	}
}

// EnvelopeStorageNG implements the storage.EnvelopeStorage interface
type EnvelopeStorageNG struct {
	model        *model.Model
	db           *yago.DB
	logger       zerolog.Logger
	eventHandler storage.EnvelopeStorageEventHandler
	tableSize    *TableSize
}

// SetEventHandler set the current handler
func (es *EnvelopeStorageNG) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	es.eventHandler = handler
}

func (es EnvelopeStorageNG) emitEvent(
	id api.UUID, status storage.EnvelopeStorageReceptionStatus, newdata bool,
) {
	if es.eventHandler != nil {
		es.eventHandler(id, status, newdata)
	}
}

// MustBegin starts a new transaction
func (es EnvelopeStorageNG) MustBegin(ctx context.Context) *RawSQLTxHelper {
	return RawSQLMustBeginContext(ctx, es.db, es.logger)
}

// GetEventTypes returns the eventtypes of an envelope, or nil if unknown
// exists is false if the envelope is unknown, true if exists.
// will panic on any SQL error.
func (es EnvelopeStorageNG) GetEventTypes(ctx context.Context, id api.UUID) (types []string, exists bool) {
	tx := es.MustBegin(ctx)
	defer tx.BlindRollback()

	validator := tools.NewEnvelopeValidatorWithQueue()

	// Load the fragments and revalidate until the eventtypes are known.
	rows, err := tx.QueryContext(ctx,
		"SELECT data FROM envelope_storage_data WHERE envelope_id = $1 AND findex > 0 ORDER BY findex",
		uuid.UUID(id))
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	for rows.Next() {
		exists = true
		var data []byte
		if err := rows.Scan(&data); err != nil {
			panic(err)
		}
		var envelope api.Envelope
		if err := proto.Unmarshal(data, &envelope); err != nil {
			panic(err)
		}

		if err := validator.Queue(&envelope); err != nil {
			panic(err)
		}

		if validator.EventTypesKnown {
			for _, ev := range validator.Events {
				types = append(types, ev.Type)
			}
			return types, true
		}
	}

	return nil, exists
}

// StoreEnvelope stores an envelope chunk and returns the envelope state in the
// storage
func (es EnvelopeStorageNG) StoreEnvelope(ctx context.Context, envelope *api.Envelope) (storage.EnvelopeState, error) {
	// If fragment is not null
	//  - calculate the fragment id
	//  - insert the fragment
	// load remaining fragments and validate
	// emit required events
	// return envelope state

	var state storage.EnvelopeState

	state.ID = envelope.GetIDAsUUID()

	var (
		nullFragment = len(envelope.EventIDs) == 0
		hasData      = false
		// TODO Get the binary form directly from the nats message
		binFragment    []byte
		fragmentID     string
		fragmentFindex int
	)

	for _, event := range envelope.Events {
		if event.Index != 0 {
			nullFragment = false
		}
		if len(event.Items) != 0 {
			hasData = true
		}
	}

	if !nullFragment {
		var err error
		binFragment, err = proto.Marshal(envelope)
		if err != nil {
			return state, err
		}
		sha := sha256.Sum256(binFragment)
		fragmentID = hex.EncodeToString(sha[:])
	}

	validator := tools.NewEnvelopeValidator()

	validatedFragmentCount := 0

	firstFragment := !nullFragment && validator.IsNext(envelope)

	if firstFragment {
		// Is it the only one ?
		if err := validator.Add(envelope); err != nil {
			return state, err
		}
		validatedFragmentCount++
		fragmentFindex = 1
	}

	tx := es.MustBegin(ctx)
	defer tx.RollbackIfOpened()

	var (
		// is the envelope already known
		newEnvelope = false

		validatorStoredState string
	)

	// if nullFragment, we won't do any writing, only read the current validator
	// state, unless it's the first fragment. If it's the fragment, we must save
	// a validator so the envelope id is known by the store
	if nullFragment {
		if _, err := tx.ExecContext(ctx,
			"SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED",
		); err != nil {
			return state, err
		}
		row := tx.QueryRowContext(ctx,
			"SELECT validator FROM envelope_storage WHERE envelope_id = $1",
			uuid.UUID(envelope.GetIDAsUUID()))
		if err := row.Scan(&validatorStoredState); err != nil {
			// TODO handle NoRows error specifically at its a user-level error
			// while other errors is unexpected internal error
			if err == sql.ErrNoRows {
				state.Status = storage.EnvelopeStorageReceptionReceiving
				newEnvelope = true
			} else {
				return state, fmt.Errorf(
					"envelope_storage_ng: could not load validator for %s: %w",
					envelope.GetIDAsUUID().String(), err)
			}
		}
	} else {
		if _, err := tx.ExecContext(ctx,
			"SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED",
		); err != nil {
			return state, err
		}
		row := tx.QueryRowContext(ctx,
			"SELECT findex, validator FROM envelope_storage WHERE envelope_id = $1 FOR UPDATE",
			uuid.UUID(envelope.GetIDAsUUID()))
		if err := row.Scan(&validatedFragmentCount, &validatorStoredState); err == sql.ErrNoRows {
			// new envelope, we'll create the record
			newEnvelope = true
		} else if err != nil {
			return state, fmt.Errorf("envelope_storage_ng: could load validator: %w", err)
		}
	}

	if validatorStoredState != "" {
		validator = tools.NewEnvelopeValidator()
		if err := jsoniter.UnmarshalFromString(validatorStoredState, &validator); err != nil {
			return state, err
		}
		sorter := tools.EnvelopeFragmentSorter{EnvelopeValidator: validator}
		sorter.Add(envelope)
		// now check for fragments that may be saved but not validated
		rows, err := tx.QueryContext(ctx,
			"SELECT fragment_id, data FROM envelope_storage_data WHERE envelope_id = $1 AND findex = 0 FOR UPDATE",
			uuid.UUID(envelope.GetIDAsUUID()))
		if err != nil {
			return state, err
		}
		var fragmentIDs []string
		var fragmentList []*api.Envelope

		if err := func() (err error) {
			defer func() {
				err = rows.Close()
			}()
			for rows.Next() {
				var (
					id   string
					data []byte
				)

				if err := rows.Scan(&id, &data); err != nil {
					return err
				}
				var fragment api.Envelope
				if err := proto.Unmarshal(data, &fragment); err != nil {
					return err
				}
				fragmentIDs = append(fragmentIDs, id)
				fragmentList = append(fragmentList, &fragment)
				sorter.Add(&fragment)
			}
			return nil
		}(); err != nil {
			return state, err
		}

		for sorter.Next() {
			fragment, err := sorter.Get()
			if err != nil {
				return state, err
			}
			validatedFragmentCount++
			if fragment == envelope {
				fragmentFindex = validatedFragmentCount
			} else {
				var found bool
				for i, f := range fragmentList {
					if f == fragment {
						found = true
						if _, err := tx.ExecContext(ctx,
							"UPDATE envelope_storage_data SET findex = $2 WHERE fragment_id = $1",
							fragmentIDs[i], validatedFragmentCount); err != nil {
							return state, err
						}
						break
					}
				}
				if !found {
					panic("the fragment returned by the sorter is unknown, which is impossible")
				}
			}
		}
		validator = sorter.EnvelopeValidator
	}

	// validator is up-to-date, we can now save the fragment and validator state
	if !nullFragment {
		q := `INSERT INTO envelope_storage_data (
		envelope_id, fragment_id, data, findex
	) VALUES (
		$1, $2, $3, $4
	)`

		// Store the fragment
		if _, err := tx.ExecContext(ctx,
			q, uuid.UUID(envelope.GetIDAsUUID()), fragmentID, binFragment, fragmentFindex,
		); err != nil {
			return state, err
		}
	}

	state.TypesKnown = validator.EventTypesKnown
	if validator.Complete {
		state.Status = storage.EnvelopeStorageReceptionComplete
	} else {
		state.Status = storage.EnvelopeStorageReceptionReceiving
	}

	for _, eventv := range validator.Events {
		eventState := storage.EventState{
			ID:             eventv.EventID,
			Type:           eventv.Type,
			AvailableItems: eventv.CurrentItemCount,
			TotalItems:     eventv.ItemCount,
			Err:            eventv.Error,
			ReadPos: storage.EventReadPos{
				Index: eventv.CurrentIndex,
			},
		}
		switch {
		case eventv.Error != nil:
			eventState.Status = storage.EnvelopeStorageReceptionError
			state.Status = storage.EnvelopeStorageReceptionError
		case eventv.Complete:
			eventState.Status = storage.EnvelopeStorageReceptionComplete
		default:
			eventState.Status = storage.EnvelopeStorageReceptionReceiving
		}
		state.Events = append(state.Events, eventState)
	}

	// if dbEnvelope.Status == storage.EnvelopeStorageReceptionStalled {
	//	dbEnvelope.Status = storage.EnvelopeStorageReceptionReceiving
	//}

	textStatus, err := state.Status.MarshalText()
	if err != nil {
		return state, err
	}

	// in any case, update the envelope timestamp && status
	query := `UPDATE envelope_storage SET updated_at = $2, status = $3, findex = $4, validator = $5 WHERE envelope_id = $1`

	if newEnvelope {
		query = `INSERT INTO envelope_storage
		(envelope_id, updated_at, status, findex, validator)
		VALUES ($1, $2, $3, $4, $5)`
	}

	if validatorStoredState, err = jsoniter.MarshalToString(validator); err != nil {
		return state, err
	}
	if _, err := tx.ExecContext(ctx, query, uuid.UUID(envelope.GetIDAsUUID()), time.Now().UTC(), string(textStatus), validatedFragmentCount, validatorStoredState); err != nil {
		return state, err
	}

	tx.MustCommit()

	es.emitEvent(state.ID, state.Status, hasData)

	return state, nil
}

// CheckStalled mark the given envelope as stalled if no fragment was
// received in the given time _and_ the envelope is in state "Receiving"
// Returns true if the envelope is indeed stalled
// If the envelope is already marked stalled, the given time is ignored.
func (es EnvelopeStorageNG) CheckStalled(ctx context.Context, id api.UUID, delay time.Duration) bool {
	tx := es.MustBegin(ctx)
	defer tx.RollbackIfOpened()

	var (
		updatedAt time.Time
		status    string
	)

	row := tx.QueryRowContext(ctx,
		`SELECT updated_at, status
		FROM envelope_storage
		WHERE envelope_id = $1`, uuid.UUID(id))
	if err := row.Scan(&updatedAt, &status); err != nil {
		if err != sql.ErrNoRows {
			return false
		}
		panic(err)
	}

	if status == "complete" || status == "error" {
		return false
	}
	if status == "stalled" {
		return true
	}
	if status == "receiving" {
		if time.Now().UTC().Sub(updatedAt) > delay {
			if _, err := tx.ExecContext(ctx,
				"UPDATE envelope_storage SET status = 'stalled' WHERE envelope_id = $1",
				uuid.UUID(id)); err != nil {
				panic(err)
			}
			tx.MustCommit()
			es.emitEvent(id, storage.EnvelopeStorageReceptionStalled, false)
			return true
		}
	}
	return false
}

func fragmentSize(fragment *api.Envelope) int {
	return proto.Size(fragment)
}

// ReceptionStatus returns the envelope reception status
func (es EnvelopeStorageNG) ReceptionStatus(ctx context.Context, id api.UUID) (storage.EnvelopeStorageReceptionStatus, error) {
	tx := es.MustBegin(ctx)
	defer tx.BlindRollback()

	var (
		status storage.EnvelopeStorageReceptionStatus
		s      string
	)

	row := tx.QueryRowContext(ctx,
		"SELECT status FROM envelope_storage WHERE envelope_id = $1", uuid.UUID(id))
	if err := row.Scan(&s); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return 0, storage.NewNoSuchIDError(storage.DatatypeEnvelope, id)
		}
		return 0, err
	}

	if err := status.UnmarshalText([]byte(s)); err != nil {
		return 0, err
	}

	return status, nil
}

// ReadEnvelope reads an envelope from the given position
func (es EnvelopeStorageNG) ReadEnvelope(ctx context.Context, id api.UUID, position storage.EnvelopeReadPos, maxSize int) (*api.Envelope, storage.EnvelopeReadPos, error) { //nolint:typecheck,staticcheck,nolintlint
	// XXX lowering maxSize will disable fragment merging, which seems to be
	// buggy.
	maxSize = 1 //nolint:typecheck,staticcheck,nolintlint

	position = position.Clone()

	tx := es.MustBegin(ctx)
	defer tx.BlindRollback()

	sorter := tools.NewEnvelopeFragmentSorter()
	sorterPosition := storage.EnvelopeStartPos()

	var (
		mergedFragment     api.Envelope
		mergedFragmentSize int
	)

	// Lookup the fragment with identical position
	rows, err := tx.QueryContext(ctx, `SELECT data FROM envelope_storage_data
	WHERE envelope_id = $1 AND findex > 0
	ORDER BY findex
	`, uuid.UUID(id))
	if err != nil {
		return nil, position, err
	}
	defer rows.Close()
	for rows.Next() {
		var data []byte
		if err := rows.Scan(&data); err != nil {
			return nil, position, err
		}
		var fragment api.Envelope
		if err := proto.Unmarshal(data, &fragment); err != nil {
			return nil, position, err
		}
		sorter.Add(&fragment)
		for sorter.Next() {
			nextPosition := sorterPosition.Clone()
			nextPosition.Start = false
			nextPosition.Complete = sorter.Complete
			for _, event := range sorter.Events {
				eventPos := nextPosition.Events[event.EventID]
				eventPos.Index = event.CurrentIndex
				nextPosition.Events[event.EventID] = eventPos
			}
			fragment, err := sorter.Get()
			if err != nil {
				return &mergedFragment, nextPosition, err
			}
			fragment.Last = sorter.Complete
			if sorterPosition.Equals(position) {
				// start merging fragments
				mergedFragment = api.Envelope{
					ID:   fragment.ID,
					Last: fragment.Last,
				}
				if err := mergedFragment.Merge(fragment); err != nil {
					return nil, position, err
				}
				mergedFragmentSize = fragmentSize(fragment)
			} else if mergedFragmentSize != 0 {
				// merge is the size permits
				fSize := fragmentSize(fragment)
				if maxSize == 0 || mergedFragmentSize+fSize < maxSize {
					if err := mergedFragment.Merge(fragment); err != nil {
						return &mergedFragment, sorterPosition, err
					}
					mergedFragmentSize = fragmentSize(&mergedFragment)
				} else {
					return &mergedFragment, sorterPosition, nil
				}
			}
			if sorter.Complete {
				return &mergedFragment, nextPosition, nil
			}
			sorterPosition = nextPosition
		}
	}
	if mergedFragmentSize != 0 {
		return &mergedFragment, sorterPosition, nil
	}

	// no fragment found, checking if the envelope actually exists
	var exists bool
	row := tx.QueryRowContext(ctx,
		"SELECT EXISTS(SELECT 1 FROM envelope_storage WHERE envelope_id = $1 LIMIT 1)",
		uuid.UUID(id),
	)
	if err := row.Scan(&exists); err != nil {
		return nil, position, err
	}

	if !exists {
		return nil, position, storage.NewNoSuchIDError(
			storage.DatatypeEnvelope, id)
	}

	return &api.Envelope{}, position, nil
}

// Purge removes all envelopes selected by the given filter
func (es EnvelopeStorageNG) Purge(ctx context.Context, filter storage.EnvelopeFilter) {
	batchSize := 1000 * 1000

	allIDs := make(chan api.UUID, batchSize*2)
	purgeIDs := make(chan []api.UUID, 2)
	errC := make(chan error, 3)

	// When purge takes a context, we'll use it
	ctx, stop := context.WithCancel(ctx)

	catchErrors := func() {
		if r := recover(); r != nil {
			stop()
			if err, ok := r.(error); ok {
				errC <- err
			} else {
				errC <- fmt.Errorf("caught panic: %v", r)
			}
		} else {
			errC <- nil
		}
	}

	// read all the envelopes in transaction-independent batches,
	go func() {
		defer catchErrors()
		defer func() { close(allIDs) }()

		offset := 0

		for func() (cont bool) {
			tx := es.MustBegin(ctx)
			defer tx.BlindRollback()

			rows, err := tx.QueryContext(ctx,
				`SELECT envelope_id FROM envelope_storage
				ORDER BY id
				LIMIT $1
				OFFSET $2`,
				batchSize*10, offset,
			)
			must(err)
			defer rows.Close()

			for rows.Next() {
				var id uuid.UUID
				must(rows.Scan(&id))
				select {
				case <-ctx.Done():
					return
				case allIDs <- api.UUID(id):
				}
				cont = true
			}

			offset += batchSize * 10
			return
		}() {
		}
	}()

	// filter with the callback
	go func() {
		defer func() {
			// purge the allIDs chan
			for range allIDs { //nolint:revive
			}
		}()
		defer catchErrors()
		defer func() { close(purgeIDs) }()

		inputBatch := make([]api.UUID, 0, batchSize)
		for {
			select {
			case <-ctx.Done():
				return
			case id, ok := <-allIDs:
				if !ok {
					if len(inputBatch) != 0 {
						select {
						case <-ctx.Done():
						case purgeIDs <- filter(inputBatch):
						}
					}
					return
				}
				inputBatch = append(inputBatch, id)
				if len(inputBatch) == batchSize {
					select {
					case <-ctx.Done():
						return
					case purgeIDs <- filter(inputBatch):
					}
					inputBatch = make([]api.UUID, 0, batchSize)
				}
			}
		}
	}()

	// remove the envelopes that where not filtered out, by batches,
	// in independent transactions

	go func() {
		defer func() {
			// purge the purgeIDs chan
			for range purgeIDs { //nolint:revive
			}
		}()
		defer catchErrors()

		var input []uuid.UUID

		removeN := func(n int) {
			tx := es.MustBegin(ctx)
			defer tx.RollbackIfOpened()

			var ids []interface{}

			for i := 0; i != n; i++ {
				ids = append(ids, input[i])
			}

			input = input[n:]

			_, err := tx.ExecContext(ctx,
				"DELETE FROM envelope_storage WHERE envelope_id = ANY ($1)",
				pq.Array(ids))
			must(err)

			_, err = tx.ExecContext(ctx,
				"DELETE FROM envelope_storage_data WHERE envelope_id = ANY ($1)",
				pq.Array(ids))
			must(err)

			tx.MustCommit()
		}

		for {
			select {
			case <-ctx.Done():
			case batch, ok := <-purgeIDs:
				if !ok {
					if len(input) != 0 {
						removeN(len(input))
					}
					return
				}
				for _, id := range batch {
					input = append(input, uuid.UUID(id))
				}
				for len(input) >= batchSize {
					if ctx.Err() != nil {
						return
					}
					removeN(batchSize)
				}
			}
		}
	}()

	err := utils.CombineErrors(
		"failed to purge. Got error(s): ", <-errC, <-errC, <-errC)
	if err != nil {
		panic(err)
	}
}

// Stat returns a summary of the storage content
func (es EnvelopeStorageNG) Stat(ctx context.Context) ([]*control.StorageStat_Entry, error) {
	tx := es.MustBegin(ctx)
	defer tx.BlindRollback()

	stat := control.StorageStat_Entry{
		Name:           "default",
		Type:           control.StorageType_ENVELOPE,
		Implementation: "pg-ng",
	}

	if err := es.tableSize.Count(tx, &stat.Count); err != nil {
		return nil, err
	}

	if err := tx.QueryRowContext(ctx,
		"SELECT pg_total_relation_size('envelope_storage') + pg_total_relation_size('envelope_storage_data')").
		Scan(&stat.EstimatedSize); err != nil {
		return nil, err
	}

	return []*control.StorageStat_Entry{&stat}, nil
}
