// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package mocks

import (
	"context"
	"fmt"
	"math"
	"sync"
	"time"

	"github.com/rs/zerolog"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/storage"
)

// NewEnvelopeStorage init an empty EnvelopeStorage suitable for tests
func NewEnvelopeStorage(logger zerolog.Logger) *EnvelopeStorage {
	return &EnvelopeStorage{
		receivers:   make(map[api.UUID]*xbus.EnvelopeReceiver),
		envelopes:   make(map[api.UUID]*api.Envelope),
		lastReceive: make(map[api.UUID]time.Time),
		isStalled:   make(map[api.UUID]bool),
		logger:      logger,
	}
}

// EnvelopeStorage implements EnvelopeStorage on top of a SQL database
type EnvelopeStorage struct {
	receivers   map[api.UUID]*xbus.EnvelopeReceiver
	envelopes   map[api.UUID]*api.Envelope
	lastReceive map[api.UUID]time.Time
	isStalled   map[api.UUID]bool
	logger      zerolog.Logger
	evHandler   storage.EnvelopeStorageEventHandler

	size uint64

	m sync.Mutex
}

// SetEventHandler sets the event handler
func (s *EnvelopeStorage) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	s.evHandler = handler
}

// GetEventTypes returns an envelop event types
func (s *EnvelopeStorage) GetEventTypes(_ context.Context, id api.UUID) ([]string, bool) {
	r, ok := s.receivers[id]
	if !ok {
		return nil, false
	}
	eventtypes := r.EventTypes()
	return eventtypes, len(eventtypes) != 0
}

func (s *EnvelopeStorage) receiveEnvelope(id api.UUID) {
	envelope, ok := s.envelopes[id]
	r := s.receivers[id]

	if !ok {
		envelope = &api.Envelope{
			ID: r.ID.ToBytes(),
		}
		s.envelopes[id] = envelope
	}

	for _, event := range r.PartialEvents() {
		index := -1
		for i, tgt := range envelope.Events {
			if tgt.GetIDAsUUID() == event.GetIDAsUUID() {
				index = i
				break
			}
		}
		if index == -1 {
			index = len(envelope.Events)
			envelope.Events = append(envelope.Events, &api.Event{})
		}

		envelope.Events[index].ID = event.ID
		envelope.Events[index].Type = event.Type
		envelope.Events[index].ItemCount = event.ItemCount
		envelope.Events[index].Checksum = event.Checksum
		envelope.Events[index].Items = append(
			envelope.Events[index].Items,
			r.NextEventItems(event.GetIDAsUUID(), int(math.MaxInt64))...)
	}

	s.envelopes[id] = envelope
}

// StoreEnvelope store a part of an envelope and returns its status
// in the storage
func (s *EnvelopeStorage) StoreEnvelope(_ context.Context, e *api.Envelope) (storage.EnvelopeState, error) {
	s.m.Lock()
	defer s.m.Unlock()

	status := storage.EnvelopeState{ID: e.GetIDAsUUID()}

	r, ok := s.receivers[e.GetIDAsUUID()]
	if !ok {
		newr, err := xbus.NewEnvelopeReceiver("", e)
		if err != nil {
			status.Status = storage.EnvelopeStorageReceptionError
			status.Err = err
			return status, err
		}
		s.receivers[e.GetIDAsUUID()] = newr
		r = newr
		s.size += uint64(16)
	} else {
		if err := r.AddFragment(e); err != nil {
			status.Status = storage.EnvelopeStorageReceptionError
			status.Err = err
			return status, err
		}
	}

	for _, ev := range e.Events {
		s.size += uint64(16 + len(ev.Type) + 16 + 4)
		for _, item := range ev.Items {
			s.size += uint64(len(item))
		}
	}

	s.isStalled[e.GetIDAsUUID()] = false
	s.lastReceive[e.GetIDAsUUID()] = time.Now()

	status.TypesKnown = r.EventTypes() != nil
	switch r.Status() {
	case api.EnvelopeAck_RECEIVING:
		status.Status = storage.EnvelopeStorageReceptionReceiving
	case api.EnvelopeAck_ACCEPTED:
		status.Status = storage.EnvelopeStorageReceptionComplete
	case api.EnvelopeAck_ERROR:
		status.Status = storage.EnvelopeStorageReceptionError
	}

	s.receiveEnvelope(e.GetIDAsUUID())

	envelope := s.envelopes[e.GetIDAsUUID()]
	events := make(map[api.UUID]*api.Event)
	for _, ev := range r.Events() {
		events[ev.GetIDAsUUID()] = ev
	}

	var newData bool

	for _, ev := range envelope.Events {
		if len(ev.Items) != 0 {
			newData = true
		}
		if knownEvent, ok := events[ev.GetIDAsUUID()]; ok && ev.ItemCount == 0 && knownEvent.ItemCount != 0 {
			ev.ItemCount = knownEvent.ItemCount
		}
		evS := storage.EventState{
			ID:             ev.GetIDAsUUID(),
			Type:           ev.Type,
			AvailableItems: uint64(len(ev.Items)),
			TotalItems:     ev.ItemCount,
			ReadPos: storage.EventReadPos{
				Index:     ev.Index,
				ItemCount: uint64(len(ev.Items)),
			},
		}
		if evS.TotalItems != 0 && evS.TotalItems == evS.AvailableItems {
			evS.Status = storage.EnvelopeStorageReceptionComplete
		} else {
			evS.Status = storage.EnvelopeStorageReceptionReceiving
		}
		status.Events = append(status.Events, evS)
	}

	if s.evHandler != nil {
		s.evHandler(status.ID, status.Status, newData)
	}

	return status, nil
}

// CheckStalled ...
func (s *EnvelopeStorage) CheckStalled(_ context.Context, id api.UUID, delay time.Duration) bool {
	s.m.Lock()
	defer s.m.Unlock()

	if _, ok := s.envelopes[id]; !ok {
		return false
	}
	if stalled := s.isStalled[id]; stalled {
		return true
	}
	lastReceive := s.lastReceive[id]
	if time.Since(lastReceive) > delay {
		s.isStalled[id] = true
		s.evHandler(id, storage.EnvelopeStorageReceptionStalled, false)
		return true
	}
	return false
}

// ReceptionStatus returns the envelope reception status
func (s *EnvelopeStorage) ReceptionStatus(_ context.Context, id api.UUID) (storage.EnvelopeStorageReceptionStatus, error) {
	if stalled := s.isStalled[id]; stalled {
		return storage.EnvelopeStorageReceptionStalled, nil
	}

	r, ok := s.receivers[id]
	if !ok {
		return storage.EnvelopeStorageReceptionUnknown, storage.NewNoSuchIDError("envelope", id)
	}
	switch r.Status() {
	case api.EnvelopeAck_RECEIVING:
		return storage.EnvelopeStorageReceptionReceiving, nil
	case api.EnvelopeAck_ACCEPTED:
		return storage.EnvelopeStorageReceptionComplete, nil
	case api.EnvelopeAck_ERROR:
		return storage.EnvelopeStorageReceptionError, nil
	default:
		return storage.EnvelopeStorageReceptionUnknown, fmt.Errorf("Invalid status: %s", r.Status())
	}
}

// ReadEnvelope reads as much as it can of an envelope, starting at 'position'
// The returned envelope should be suitable for encrypted nats transfert, ie
// its total jsonified size should be under maxsize
func (s *EnvelopeStorage) ReadEnvelope(_ context.Context, id api.UUID, position storage.EnvelopeReadPos, maxsize int) (*api.Envelope, storage.EnvelopeReadPos, error) {
	s.m.Lock()
	defer s.m.Unlock()

	envelope, ok := s.envelopes[id]
	if !ok {
		return nil, position, storage.NewNoSuchIDError("envelope", id)
	}

	if position.Start {
		for _, e := range envelope.Events {
			position.Events[e.GetIDAsUUID()] = storage.EventReadPos{}
		}
		position.Start = false
	}
	fragment := api.Envelope{
		ID: envelope.ID,
	}
	for _, event := range envelope.Events {
		fragment.EventIDs = append(fragment.EventIDs, event.ID)
	}

	// TODO in case we already know the envelope is small enough, return
	// it complete at once

	// approx size of the jsonified envelope before any event is added
	estimateSize := 36 + len(fragment.EventIDs)*3

	position = position.Clone()

	tgtEvents := map[api.UUID]*api.Event{}

	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.GetIDAsUUID()]
		if eventPos.ItemCount < uint64(len(ev.Items)) {
			eventPos.Index++
			tgtEvents[ev.GetIDAsUUID()] = &api.Event{
				ID:        ev.ID,
				Type:      ev.Type,
				Index:     eventPos.Index,
				ItemCount: ev.ItemCount,
				Checksum:  ev.Checksum,
			}
			estimateSize += 36 + len(ev.Type) + 8
		}
		position.Events[ev.GetIDAsUUID()] = eventPos
	}

	for maxsize == 0 || estimateSize < maxsize {
		before := estimateSize
		for _, ev := range envelope.Events {
			pos := position.Events[ev.GetIDAsUUID()]
			if pos.ItemCount == uint64(len(ev.Items)) {
				continue
			}
			tgt := tgtEvents[ev.GetIDAsUUID()]
			item := ev.Items[pos.ItemCount]
			tgt.Items = append(tgt.Items, item)
			pos.ItemCount++

			estimateSize += int(float64(len(item)) * 1.1)

			position.Events[ev.GetIDAsUUID()] = pos
			tgtEvents[ev.GetIDAsUUID()] = tgt
			if maxsize != 0 && estimateSize >= maxsize {
				break
			}
		}
		if estimateSize == before {
			break
		}
	}

	for _, ev := range tgtEvents {
		fragment.Events = append(fragment.Events, ev)
	}

	complete := true
	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.GetIDAsUUID()]
		if ev.ItemCount == 0 || eventPos.ItemCount != ev.ItemCount {
			complete = false
		}
	}

	if complete {
		fragment.Last = true
		position.Complete = true
	}

	return &fragment, position, nil
}

// Purge removes the envelopes fitered by the given filter
func (s *EnvelopeStorage) Purge(_ context.Context, filter storage.EnvelopeFilter) {
	var todelete []api.UUID

	for id := range s.envelopes {
		todelete = append(todelete, filter([]api.UUID{id})...)
	}

	for _, id := range todelete {
		delete(s.envelopes, id)
		delete(s.receivers, id)
	}
}

// Stat ...
func (s *EnvelopeStorage) Stat(_ context.Context) ([]*control.StorageStat_Entry, error) {
	stat := control.StorageStat_Entry{
		Count:         uint64(len(s.envelopes)),
		EstimatedSize: s.size,
	}

	return []*control.StorageStat_Entry{&stat}, nil
}
