// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package mocks

import (
	"context"
	"fmt"
	"time"

	"xbus.io/xbus/storage"
)

// SessionStorage is a in-memory implementation of storage.SessionStorage
type SessionStorage struct {
	sessions []storage.Session
}

// List returns all the opened sessions
func (store *SessionStorage) List(context.Context) ([]storage.Session, error) {
	return store.sessions, nil
}

// Get returns a Session from its token
func (store *SessionStorage) Get(_ context.Context, token string) (*storage.Session, error) {
	for _, session := range store.sessions {
		if session.Token == token {
			return &session, nil
		}
	}
	return nil, fmt.Errorf("%w: session '%s'", storage.ErrNotExist, token)
}

// Set create or update a Session
func (store *SessionStorage) Set(_ context.Context, session storage.Session) error {
	for i, s := range store.sessions {
		if s.Token == session.Token {
			store.sessions[i] = session
			return nil
		}
	}
	store.sessions = append(store.sessions, session)
	return nil
}

// Delete deletes a token
func (store *SessionStorage) Delete(_ context.Context, token string) error {
	for i, session := range store.sessions {
		if session.Token == token {
			store.sessions = append(
				store.sessions[:i], store.sessions[i+1:]...)
			return nil
		}
	}
	return nil
}

// DeleteOlderThan deletes the sessions based on their 'validUntil'
func (store *SessionStorage) DeleteOlderThan(_ context.Context, time time.Time) error {
	var sessions []storage.Session
	for _, session := range store.sessions {
		if session.ValidUntil.AsTime().After(time) {
			sessions = append(sessions, session)
		}
	}
	store.sessions = sessions
	return nil
}
