// Copyright 2017 Orus.io. All rights reserved.

package xbus

import (
	"context"
	"fmt"
	"time"

	nats "github.com/nats-io/nats.go"
	"github.com/nats-rpc/nrpc"
	"github.com/rs/zerolog"
	"google.golang.org/protobuf/types/known/timestamppb"

	api "xbus.io/go-xbus/v4/api"
)

// EnvelopeSenderDefaultDelay is the default MaxDelay of the envelope senders
// created by Actor.Output()
var EnvelopeSenderDefaultDelay = 10 * time.Second

func catchErrors(log zerolog.Logger) {
	if r := recover(); r != nil {
		log.Error().Stack().Interface("error", r).Msg("Caught error")
	}
}

// ActorProcessHandler is a function that implements a Actor.Process
type ActorProcessHandler func(*ActorProcessingContext) error

// newActor initialize an actor
func newActor(client *Client, options ActorOptions) *Actor {
	actor := Actor{
		ID:      api.MustUUIDFromBytes(options.ID),
		Name:    options.Name,
		Kind:    options.Kind,
		Options: options,
		Client:  client,
	}
	logContext :=
		client.Log.With().
			Str("actor", actor.Name).
			Str("actor_id", actor.ID.String())
	if options.Service.Type != "" {
		logContext = logContext.Str("service_type", options.Service.Type)
	}
	actor.Log = NewLogger(logContext.Logger())
	return &actor
}

// Actor provides the common tools for all actors
type Actor struct {
	ID      api.UUID
	Name    string
	Kind    api.Actor_Kind
	Options ActorOptions
	Log     Logger
	Client  *Client
}

// GetClient returns the underlying Client instance
func (a *Actor) GetClient() *Client {
	return a.Client
}

// ProcessingEnd calls the ActorProcessingEnd xbus api
func (a *Actor) ProcessingEnd(
	context *api.ProcessingContext,
	status api.ActorProcessingState_Status,
	messages []*api.LogMessage,
) error {
	_, err := a.GetClient().tryProcessingEnd(
		a.ID, context, status, messages, true)

	return err
}

// ProcessingSuccess signal the success of a processing
// for actors with no outputs of it not all outputs were
// hit or closed
func (a *Actor) ProcessingSuccess(context *api.ProcessingContext) error {
	return a.ProcessingEnd(context, api.ActorProcessingState_SUCCESS, nil)
}

// ProcessingError signals a processing error.
func (a *Actor) ProcessingError(context *api.ProcessingContext, format string, fmtArgs ...interface{}) error {
	return a.ProcessingEnd(context, api.ActorProcessingState_ERROR, []*api.LogMessage{{
		Time:  timestamppb.Now(),
		Level: api.LogLevel_ERROR,
		Text:  fmt.Sprintf(format, fmtArgs...),
	}})
}

// SendEnvelope send a potentially incomplete envelope to the broker
func (a *Actor) SendEnvelope(
	ctx *api.ProcessingContext, output string, env *api.Envelope,
) (status api.EnvelopeAck_ReceptionStatus, err error) {
	if a.Kind == api.Actor_CONSUMER {
		return api.EnvelopeAck_ERROR, fmt.Errorf("A 'Consumer' actor cannot emit to outputs")
	}
	return a.Client.SendEnvelopeFragment(a.ID, ctx, output, env)
}

// SendItems send a group of items to the broker
// The envelope and event will be constructed automatically
func (a *Actor) SendItems(
	ctx *api.ProcessingContext, output, eventType string, items []api.Item,
) (status api.EnvelopeAck_ReceptionStatus, err error) {
	return a.SendEnvelope(ctx, output, ForgeEnvelopeFromItems(eventType, items))
}

// ForgeEnvelopeFromEvent ...
func ForgeEnvelopeFromEvent(event *api.Event) *api.Envelope {
	envID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	return &api.Envelope{
		ID:       envID.ToBytes(),
		EventIDs: [][]byte{event.ID},
		Events:   []*api.Event{event},
	}
}

// ForgeEnvelopeFromItems ...
func ForgeEnvelopeFromItems(eventType string, items []api.Item) *api.Envelope {
	evID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	checksum := api.Checksum(0).
		Update(evID[:16]).
		Update([]byte(eventType))

	for _, item := range items {
		checksum = checksum.Update(item)
	}

	event := api.Event{
		ID:        evID.ToBytes(),
		Type:      eventType,
		Index:     1,
		Items:     api.ItemsAsChunks(items),
		ItemCount: uint64(len(items)),
		Checksum:  checksum.AsUint32(),
	}
	return ForgeEnvelopeFromEvent(&event)
}

type actorAgentServer struct {
	actor  *Actor
	handle ActorProcessHandler
}

func (s actorAgentServer) Process(
	context context.Context, req *api.ActorProcessRequest,
	send func(*api.ActorProcessingState),
) error {
	log := NewLogger(s.actor.Log.With().Object("processing_context", req.Context).Logger())
	apc := ActorProcessingContext{
		Request: req,
		Context: context,
		Actor:   s.actor,
		Log:     log,
		currentState: &api.ActorProcessingState{
			Context: req.Context,
			Status:  api.ActorProcessingState_RUNNING,
		},
		send:            send,
		receiveEnvelope: s.actor.Client.ReceiveEnvelope,
		closeOutput:     s.actor.CloseOutput,
		sendFragment:    s.actor.SendEnvelope,
		processingEnd:   s.actor.ProcessingEnd,
	}
	if err := s.handle(&apc); err != nil {
		if apc.done {
			// The handler returned an error _and_ called Done() or Error().
			log.Error().Msgf(
				"the handler returned an error after calling Error() or Done(), "+
					"which is a forbidden behavior. Please check the handler "+
					"implementation. The state that was sent is: %v, and the "+
					"returned error is '%s'", apc.currentState, err,
			)
		} else if apc.detached {
			// The handler returned an error _and_ called Detached()
			log.Error().Msgf(
				"the handler returned an error after calling Detached(), "+
					"which is a forbidden behavior. Please check the handler "+
					"implementation. The returned error is '%s'", err,
			)
		} else if _, ok := err.(*nrpc.Error); ok {
			return err
		} else {
			apc.Error(err)
		}
	} else if !apc.detached && !apc.done {
		apc.Done()
	}
	return nil
}

// Subscription is the actor subscription for handling incoming process requests
type Subscription struct {
	processSub *nats.Subscription
	surveySub  *nats.Subscription
	actor      *Actor
	workers    *nrpc.WorkerPool
}

func (s Subscription) handleSurvey(request *api.SurveyRequest) {
	defer catchErrors(s.actor.Log.Logger)

	if request.OnlineStatus {
		if err := s.actor.Client.API.Sentinel.ActorReady(s.actor.ID.String()); err != nil {
			s.actor.Log.Err(err).Msg("Error calling 'ActorReady'")
		}
	}
}

// Unsubscribe the subscription
func (s Subscription) Unsubscribe() error {
	if err := s.actor.Client.API.Sentinel.ActorLeaving(s.actor.ID.String()); err != nil {
		s.actor.Log.Err(err).Msg("Error calling 'ActorLeaving'")
	}
	s.actor.Log.Info().Msg("ActorLeaving called")
	err, err2 := s.processSub.Unsubscribe(), s.surveySub.Unsubscribe()
	if err == nil {
		err = err2
	}
	s.workers.Close(time.Second)
	return err
}

// Subscribe registers a ActorAgent implementation and starts listening
func (a *Actor) Subscribe(
	handle ActorProcessHandler,
) (*Subscription, error) {
	conn := a.Client.GetConn()

	workers := nrpc.NewWorkerPool(
		context.Background(), 1, 0, time.Second)

	handler := api.NewActorAgentConcurrentHandler(
		workers,
		conn,
		actorAgentServer{
			actor:  a,
			handle: handle,
		},
	)
	// TODO keep track of active calls so the actor cannot be shutdown
	// until all current calls are canceled or ended (we should use context
	// for the canceling)
	sub, err := conn.QueueSubscribe(
		api.XbusAPIPrefix+"actoragent."+a.ID.String()+".>",
		a.ID.String(),
		handler.Handler,
	)
	if err != nil {
		workers.Close(0)
		return nil, err
	}
	subscription := &Subscription{
		processSub: sub,
		actor:      a,
		workers:    workers,
	}

	if subscription.surveySub, err = a.Client.API.Sentinel.SurveySubscribe(
		subscription.handleSurvey,
	); err != nil {
		_ = sub.Unsubscribe()
		return nil, err
	}

	if err := a.Client.API.Sentinel.ActorReady(a.ID.String()); err != nil {
		_ = sub.Unsubscribe()
		_ = subscription.surveySub.Unsubscribe()
		return nil, err
	}

	return subscription, nil
}

// CloseOutput closes an output
func (a *Actor) CloseOutput(
	context *api.ProcessingContext, output string,
) (*api.EnvelopeAck, error) {
	return a.Client.sendEnvelopeFragment(a.ID, context, output, nil)
}

// LegacyOpenOutput returns a EnvelopeSender
func (a *Actor) LegacyOpenOutput(
	context *api.ProcessingContext, output string, eventTypes ...string,
) *EnvelopeSender {
	es := NewEnvelopeSender(
		eventTypes, api.EnvelopeMaxSize,
		func(env *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			return a.SendFragment(context, output, env)
		})
	es.SetMaxDelay(EnvelopeSenderDefaultDelay)
	es.FreezeEventList()
	return es
}
