package api

import (
	"errors"
	"fmt"
	"sync"
)

var (
	ErrDuplicateID   = errors.New("duplicate id")
	ErrClosedMessage = errors.New("message closed")
	ErrLastMessage   = errors.New("last message")
	ErrNoSuchMessage = errors.New("no such message")
	ErrChecksum      = errors.New("checksum error")
)

type FragmentsBuilderOptions struct {
	MaxSize int
}

type FragmentsBuilderOptionSetter func(*FragmentsBuilderOptions)

func NewFragmentsBuilder(
	id UUID, pushFragment func(*Envelope),
	optionSetters ...FragmentsBuilderOptionSetter,
) *FragmentsBuilder {
	var options = FragmentsBuilderOptions{
		MaxSize: EnvelopeMaxSize,
	}
	for _, setter := range optionSetters {
		setter(&options)
	}
	return &FragmentsBuilder{
		id:           id,
		pushFragment: pushFragment,
		maxSize:      options.MaxSize,
	}
}

func FragmentsBuilderMaxSize(maxSize int) FragmentsBuilderOptionSetter {
	return func(options *FragmentsBuilderOptions) {
		options.MaxSize = maxSize
	}
}

type messageState struct {
	id         UUID
	msgType    string
	chunkCount uint64
	checksum   Checksum
	closed     bool
	index      uint64
}

type FragmentsBuilder struct {
	id           UUID
	pushFragment func(*Envelope)

	messages    []*messageState
	lastMessage bool

	nextFragment        *Envelope
	nextFragmentSize    int
	nextFragmentIsEmpty bool

	maxSize int
}

func (b *FragmentsBuilder) getMessageState(id UUID) (*messageState, bool) {
	for _, state := range b.messages {
		if state.id.Equal(id) {
			return state, true
		}
	}
	return nil, false
}

// AddMessage adds a message
func (b *FragmentsBuilder) AddMessage(id UUID, msgType string) error {
	if b.lastMessage {
		return ErrLastMessage
	}
	if _, ok := b.getMessageState(id); ok {
		return fmt.Errorf("%w: %s", ErrDuplicateID, id)
	}
	b.messages = append(b.messages,
		&messageState{
			id:       id,
			msgType:  msgType,
			checksum: Checksum(0).Update(id.ToBytes()).Update([]byte(msgType)),
		})

	if _, err := b.ensureEvent(id); err != nil {
		return err
	}
	return nil
}

func (b *FragmentsBuilder) getNextFragment() *Envelope {
	if b.nextFragment == nil {
		b.nextFragment = &Envelope{ID: b.id.ToBytes()}
		b.nextFragmentSize = 16 + 1
		b.nextFragmentIsEmpty = true
	}
	return b.nextFragment
}

// LastMessage signals that no more message will be added
func (b *FragmentsBuilder) LastMessage() {
	nextFragment := b.getNextFragment()

	if b.flushIfNeeded(len(b.messages)*17 + 1) {
		nextFragment = b.getNextFragment()
	}
	for _, msg := range b.messages {
		nextFragment.EventIDs = append(nextFragment.EventIDs, msg.id.ToBytes())
		b.nextFragmentSize += len(b.messages)*17 + 1
	}

	b.nextFragmentIsEmpty = false
	b.lastMessage = true

	for _, msg := range b.messages {
		if !msg.closed {
			return
		}
	}

	// We want eventIDs to be sent as soon as possible so the other side knows
	// what to expect
	b.flush()
}

func (b *FragmentsBuilder) getMessageAndEvent(id UUID) (*messageState, *Event, error) {
	message, ok := b.getMessageState(id)
	if !ok {
		return nil, nil, fmt.Errorf("%w: %s", ErrNoSuchMessage, id)
	}

	if message.closed {
		return nil, nil, ErrClosedMessage
	}

	nextFragment := b.getNextFragment()

	var event *Event
	for _, e := range nextFragment.Events {
		if e.GetIDAsUUID().Equal(id) {
			event = e
		}
	}

	return message, event, nil
}

func (b *FragmentsBuilder) ensureEvent(id UUID) (*Event, error) {
	message, ok := b.getMessageState(id)
	if !ok {
		return nil, fmt.Errorf("%w: %s", ErrNoSuchMessage, id)
	}

	nextFragment := b.getNextFragment()

	var event *Event
	for _, e := range nextFragment.Events {
		if e.GetIDAsUUID().Equal(id) {
			event = e
		}
	}

	if event == nil {
		eventSize := 17 + len(message.msgType) + 1 + 4
		if b.flushIfNeeded(eventSize) {
			nextFragment = b.getNextFragment()
		}
		message.index++
		event = &Event{ID: id.ToBytes(), Type: message.msgType, Index: message.index}
		nextFragment.Events = append(nextFragment.Events, event)
		b.nextFragmentSize += eventSize
	}

	return event, nil
}

func (b *FragmentsBuilder) getMessageAndEnsureEvent(id UUID) (*messageState, *Event, error) {
	message, event, err := b.getMessageAndEvent(id)
	if err != nil {
		return nil, nil, err
	}
	if event == nil {
		e, err := b.ensureEvent(id)
		if err != nil {
			return nil, nil, err
		}
		event = e
	}
	return message, event, nil
}

// flushIfNeeded flushes if the added size makes the next fragment bigger that
// the maxSize.
// returns true if flushed occured
func (b *FragmentsBuilder) flushIfNeeded(addedsize int) bool {
	if b.nextFragment != nil && (b.nextFragmentSize+addedsize) > b.maxSize {
		b.flush()
		return true
	}
	return false
}

// AddChunk adds a chunk to an event.
func (b *FragmentsBuilder) AddChunk(id UUID, chunk []byte) error {
	message, event, err := b.getMessageAndEvent(id)
	if err != nil {
		return err
	}

	if b.flushIfNeeded(len(chunk) + 16 + len(message.msgType)) {
		message, event, err = b.getMessageAndEvent(id)
		if err != nil {
			return err
		}
	}

	message.chunkCount++
	message.checksum = message.checksum.Update(chunk)

	if event == nil {
		event, err = b.ensureEvent(id)
		if err != nil {
			return err
		}
	}

	event.Items = append(event.Items, chunk)
	b.nextFragmentSize += len(chunk) + 16 + len(message.msgType)
	b.nextFragmentIsEmpty = false
	return nil
}

func (b *FragmentsBuilder) CloseMessage(id UUID, checksum Checksum) error {
	message, event, err := b.getMessageAndEnsureEvent(id)
	if err != nil {
		return err
	}
	if checksum != 0 {
		if checksum != message.checksum {
			return fmt.Errorf("%w: expected 0x%x, got 0x%x", ErrChecksum, message.checksum, checksum)
		}
	}
	event.Checksum = message.checksum.AsUint32()
	event.ItemCount = message.chunkCount
	b.nextFragmentSize += 8

	b.nextFragmentIsEmpty = false
	message.closed = true

	if !b.lastMessage {
		return nil
	}
	for _, msg := range b.messages {
		if !msg.closed {
			return nil
		}
	}
	b.nextFragment.Last = true
	b.flush()
	return nil
}

func (b *FragmentsBuilder) flush() {
	if !b.nextFragmentIsEmpty {
		b.pushFragment(b.nextFragment)
		b.nextFragment = nil
	}
}

func NewSyncFragmentsBuilder(id UUID, pushFragment func(*Envelope), setters ...FragmentsBuilderOptionSetter) *SyncFragmentsBuilder {
	return &SyncFragmentsBuilder{
		FragmentsBuilder: NewFragmentsBuilder(id, pushFragment, setters...),
	}
}

type SyncFragmentsBuilder struct {
	*FragmentsBuilder
	m sync.Mutex
}

// AddMessage adds a message
func (b *SyncFragmentsBuilder) AddMessage(id UUID, msgType string) error {
	b.m.Lock()
	defer b.m.Unlock()

	return b.FragmentsBuilder.AddMessage(id, msgType)
}

// LastMessage signals that no more message will be added
func (b *SyncFragmentsBuilder) LastMessage() {
	b.m.Lock()
	defer b.m.Unlock()

	b.FragmentsBuilder.LastMessage()
}

// AddChunk adds a chunk to an event.
func (b *SyncFragmentsBuilder) AddChunk(id UUID, chunk []byte) error {
	b.m.Lock()
	defer b.m.Unlock()

	return b.FragmentsBuilder.AddChunk(id, chunk)
}

func (b *SyncFragmentsBuilder) CloseMessage(id UUID, checksum Checksum) error {
	b.m.Lock()
	defer b.m.Unlock()

	return b.FragmentsBuilder.CloseMessage(id, checksum)
}
