package cmd

import (
	"context"
	"fmt"
	"os"
	"syscall"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/utils"
)

var requestCmd = cobra.Command{
	Use:   "request",
	Short: "Run a request with an emitter and wait for the response",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) == 0 || len(args) > 2 {
			if err := cmd.Usage(); err != nil {
				fmt.Println("Error printing usage:", err)
			}
			os.Exit(1)
		}
		emitter := args[0]
		input := cmd.Flag("input").Value.String()
		output := cmd.Flag("output").Value.String()
		indent := cmd.Flag("indent").Value.String() == "true"
		os.Exit(doRequest(emitter, input, output, indent))
	},
}

func doRequest(emitter, input, output string, indent bool) int {
	ctx, cancel := utils.CtxWithSignal(context.Background(), os.Interrupt, syscall.SIGTERM)
	defer cancel()
	client, err := GetClient(false)
	if err != nil {
		Logger.Err(err).Msg("Error initializing")
		return 1
	}

	if err := client.Startup(false); err != nil {
		Logger.Err(err).Msg("Error starting client")
		return 1
	}
	defer client.Shutdown()

	var actor *xbus.Actor
	if actorID, err := api.UUIDFromString(emitter); err == nil {
		actor = client.GetActor(actorID)
	} else {
		actor = client.GetActorByName(emitter)
	}
	if actor == nil {
		Logger.Err(err).Msg("No such emitter")
		return 1
	}

	request, err := loadEnvelope(input)
	if err != nil {
		Logger.Err(err).Msg("Error loading envelope")
		return 1
	}

	response, err := actor.Request(ctx, request)
	if err != nil {
		Logger.Err(err).Msg("Request, or waiting for the response, failed")
		return 1
	}

	if err := saveEnvelope(ctx, response, output, indent); err != nil {
		Logger.Err(err).Msg("could not save response")
		return 1
	}
	if output != "-" {
		Logger.Info().Msgf("Saved response to %s", output)
	}

	return 0
}

func mustUUIDV4() api.UUID {
	id, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	return id
}

func init() {
	flags := requestCmd.PersistentFlags()
	flags.StringP("input", "i", "-", "input file. Default is stdin ('-')")
	flags.StringP("output", "o", "-", "output file. Default is stdout ('-')")
	flags.Bool("indent", false, "indent the output file")
	RootCmd.AddCommand(&requestCmd)
}
