// Copyright 2012-2015 Apcera Inc. All rights reserved.
// Copyright 2016 Orus.io for the modified parts
// this file is covered by the Apache 2.0 licence

// Package xbus provides a few things that all xbus actors (services & clients)
// may need
package xbus

import (
	"bytes"
	"errors"
	"fmt"
	"io"
	"os"
	"strings"
	"time"

	jsoniter "github.com/json-iterator/go"
	"github.com/rs/zerolog"
	"google.golang.org/protobuf/types/known/timestamppb"
	api "xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"
)

var ErrNoProcessingContext = errors.New("no processing context")

// LoggerOptions is the struct you need to pass to the GetLogger function
// it can be just created and passed untouched to the GetLogger if you need
// some quick logging functionality.
// it is recommended to at least set the Debug and Trace flags according to
// your taste though
type LoggerOptions struct {
	AppName string
	Logtime bool
	Debug   bool
	Trace   bool

	// FinalOutput is where the logs are written
	FinalOutput      io.Writer
	OutputMiddleware func(io.Writer) io.Writer
}

func autoColors() bool {
	stat, err := os.Stdout.Stat()
	if err != nil || (stat.Mode()&os.ModeCharDevice) == 0 {
		return false
	}
	return true
}

// GetLogger is a convenience function that returns a logger configured
// according to your options and ready to be used in your server
func GetLogger(opts *LoggerOptions) (zerolog.Logger, io.Writer) {
	var out io.Writer = opts.FinalOutput
	if out == nil {
		out = os.Stderr
	}

	if autoColors() {
		out = zerolog.ConsoleWriter{Out: out}
	}

	if opts.OutputMiddleware != nil {
		out = opts.OutputMiddleware(out)
	}

	context := zerolog.
		New(out).
		With().
		Str("app", opts.AppName)
	if opts.Logtime {
		context = context.
			Timestamp()
	}
	log := context.Logger()
	if opts.Trace {
		log = log.Level(zerolog.TraceLevel)
	} else if opts.Debug {
		log = log.Level(zerolog.DebugLevel)
	} else {
		log = log.Level(zerolog.InfoLevel)
	}

	return log, out
}

// NewPrefixedLogger PrefixedLogger constructor
func NewPrefixedLogger(prefix string, logger zerolog.Logger) zerolog.Logger {
	return logger.With().Str("prefix", prefix).Logger()
}

// NewNoopLogger returns a NoopLogger
func NewNoopLogger() zerolog.Logger {
	return zerolog.New(io.Discard)
}

func NewLogger(log zerolog.Logger) Logger {
	return Logger{log}
}

type Logger struct {
	zerolog.Logger
}

func (log Logger) AsZero() zerolog.Logger {
	return log.Logger
}

func (log Logger) NoXbusLogger() zerolog.Logger {
	return log.Logger.With().Bool("xbus-log", false).Logger()
}

func (log Logger) NoLocalLogger() zerolog.Logger {
	return log.Logger.With().Bool("local-log", false).Logger()
}

func (log Logger) Info() LoggerEvent {
	return LoggerEvent{log.Logger.Info()}
}

func (log Logger) Warn() LoggerEvent {
	return LoggerEvent{log.Logger.Warn()}
}

func (log Logger) Error() LoggerEvent {
	return LoggerEvent{log.Logger.Error()}
}

func (log Logger) Err(err error) LoggerEvent {
	return LoggerEvent{log.Logger.Err(err)}
}

func (log Logger) Fatal() LoggerEvent {
	return LoggerEvent{log.Logger.Fatal()}
}

func (log Logger) Panic() LoggerEvent {
	return LoggerEvent{log.Logger.Panic()}
}

type LoggerEvent struct {
	*zerolog.Event
}

func (event LoggerEvent) NoXbusLog() *zerolog.Event {
	return event.Bool("xbus-log", false)
}

func (event LoggerEvent) NoLocalLog() *zerolog.Event {
	return event.Bool("local-log", false)
}

type DirectorLogFunc func(actorID string, req *api.ActorLogRequest) error

func NewXbusLogWriter(
	directorLog DirectorLogFunc, next io.Writer, logger zerolog.Logger,
) io.Writer {
	lw, ok := next.(zerolog.LevelWriter)
	if !ok {
		lw = levelWriterAdapter{next}
	}
	return xbusLogWriter{
		xbusLog: directorLog,
		logger:  logger,
		next:    lw,
	}
}

type xbusLogWriter struct {
	xbusLog func(actorID string, req *api.ActorLogRequest) error
	logger  zerolog.Logger
	next    zerolog.LevelWriter
}

func (w xbusLogWriter) Write(p []byte) (n int, err error) {
	return w.WriteLevel(tools.GetZerologMessageLevel(p), p)
}

// WriteLevel implements zerolog.LevelWriter interface
func (w xbusLogWriter) WriteLevel(level zerolog.Level, p []byte) (int, error) {
	var localLog = true
	if level >= zerolog.InfoLevel {
		localLog = hasLocalLog(p)
		if hasXbusLog(p) {
			if err := w.sendToXbus(level, p); err != nil {
				if err == ErrNoProcessingContext {
					localLog = true
				} else {
					w.logger.Err(err).Msg("could not send log to xbus")
				}
			}
		}
	}

	if localLog {
		return w.next.WriteLevel(level, p)
	}
	return 0, nil
}

func (w xbusLogWriter) sendToXbus(level zerolog.Level, p []byte) error {
	if w.xbusLog == nil {
		return nil
	}
	if level < zerolog.InfoLevel {
		return fmt.Errorf("log level %s too low for sending to xbus", level)
	}
	iter := jsoniter.ConfigFastest.BorrowIterator(p)
	defer jsoniter.ConfigFastest.ReturnIterator(iter)

	if iter.WhatIsNext() != jsoniter.ObjectValue {
		return nil
	}

	var (
		apiLevel          api.LogLevel
		timestamp         = time.Now()
		text              string
		actorID           api.UUID
		envelopeID        api.UUID
		processingContext *api.ProcessingContext
		extra             []string
	)

	switch level {
	case zerolog.InfoLevel:
		apiLevel = api.LogLevel_NOTICE
	case zerolog.WarnLevel:
		apiLevel = api.LogLevel_WARNING
	default:
		apiLevel = api.LogLevel_ERROR
	}

	if !iter.ReadObjectCB(func(iter *jsoniter.Iterator, field string) bool {
		switch field {
		case "message":
			text = iter.ReadString()
		case "time":
			t, err := time.Parse(time.RFC3339, iter.ReadString())
			if err != nil {
				iter.ReportError("parse time", err.Error())
			} else {
				timestamp = t
			}
		case "actor_id":
			value := iter.ReadString()
			if iter.Error == nil && value != "" {
				id, err := api.UUIDFromString(value)
				if err != nil {
					iter.ReportError("parse uuid", err.Error())
				}
				actorID = id
			}
		case "envelope_id", "envelope-id":
			value := iter.ReadString()
			if iter.Error == nil && value != "" {
				id, err := api.UUIDFromString(value)
				if err != nil {
					iter.ReportError("parse uuid", err.Error())
				}
				envelopeID = id
			}
		case "processing_context":
			processingContext = &api.ProcessingContext{}
			return iter.ReadObjectCB(func(iter *jsoniter.Iterator, field string) bool {
				switch field {
				case "process_id":
					value := iter.ReadString()
					if iter.Error == nil && value != "" {
						id, err := api.UUIDFromString(value)
						if err != nil {
							iter.ReportError("parse uuid", err.Error())
						}
						processingContext.ProcessID = id.ToBytes()
					}
				case "node_id":
					processingContext.NodeID = iter.ReadString()
				default:
					iter.Skip()
				}
				return iter.Error == nil
			})
		default:
			value := iter.ReadAny().ToString()
			extra = append(extra, field+"="+value)
		}
		return iter.Error == nil
	}) {
		if iter.Error != nil {
			return fmt.Errorf("could not parse zerolog event: %w", iter.Error)
		}
	}

	if len(extra) != 0 {
		text = strings.Join(extra, " ") + " " + text
	}

	if processingContext == nil {
		return ErrNoProcessingContext
	}
	if processingContext != nil {
		if err := w.addLogMessage(
			timestamp, actorID, processingContext, envelopeID, apiLevel, text,
		); err != nil {
			return err
		}
	}

	return nil
}

func (w xbusLogWriter) addLogMessage(
	timestamp time.Time, actorID api.UUID,
	context *api.ProcessingContext,
	envelopeID api.UUID, level api.LogLevel, text string,
) error {
	var req = api.ActorLogRequest{
		Context:    context,
		EnvelopeID: envelopeID.ToBytes(),
		Messages: []*api.LogMessage{{
			Time:  timestamppb.Now(),
			Level: level,
			Text:  text,
		}},
	}
	return w.xbusLog(actorID.String(), &req)
}

var (
	xbusLogMarker    = []byte(`"xbus-log":`)
	xbusLogMarkerLen = len(xbusLogMarker)

	localLogMarker    = []byte(`"local-log":`)
	localLogMarkerLen = len(localLogMarker)
)

func hasXbusLog(p []byte) bool {
	if i := bytes.Index(p, xbusLogMarker); i != -1 {
		first := p[i+xbusLogMarkerLen]
		switch first {
		case 't':
			return true
		case 'f':
			return false
		}
	}
	return true
}

func hasLocalLog(p []byte) bool {
	if i := bytes.Index(p, localLogMarker); i != -1 {
		first := p[i+localLogMarkerLen]
		switch first {
		case 't':
			return true
		case 'f':
			return false
		}
	}
	return true
}

type levelWriterAdapter struct {
	io.Writer
}

func (lw levelWriterAdapter) WriteLevel(l zerolog.Level, p []byte) (n int, err error) {
	return lw.Write(p)
}
