package xbus

import (
	"errors"
	"fmt"
	"strconv"
	"strings"
)

var (
	// ErrSettings is any setting error
	ErrSettings = errors.New("settings error")

	// ErrNoSuchKey is a err kind returned if the key is not present in the settings
	ErrNoSuchKey = fmt.Errorf("%w: no such key", ErrSettings)

	// ErrInvalidType is a err kind returnd if the setting value does not have the expected type
	ErrInvalidType = fmt.Errorf("%w: invalid type", ErrSettings)
)

func NoSuchKeyError(key string) error {
	return fmt.Errorf("%w: %s", ErrNoSuchKey, key)
}

// InvalidTypeError is returned if the value has the wrong type
type InvalidTypeError struct {
	Key          string
	ExpectedType string
	Value        interface{}
}

func (e InvalidTypeError) Error() string {
	return fmt.Sprintf("invalid type/value for key '%s'. Expected a %s, got a %T: %v",
		e.Key, e.ExpectedType, e.Value, e.Value)
}

func (e InvalidTypeError) Unwrap() error {
	return ErrInvalidType
}

// Settings holds a Actor service settings
type Settings map[string]interface{}

// Get returns a setting value
func (s Settings) Get(key string) (interface{}, bool) {
	value, ok := s[key]
	return value, ok
}

// String returns a string setting value
func (s Settings) String(key string) (string, error) {
	value, ok := s.Get(key)
	if !ok {
		return "", NoSuchKeyError(key)
	}
	if s, ok := value.(string); ok {
		return s, nil
	}
	return "", InvalidTypeError{key, "string", value}
}

// MustString ...
func (s Settings) MustString(key string) string {
	value, err := s.String(key)
	if err != nil {
		panic(err)
	}
	return value
}

// StringD returns a string setting value, with a default value if the setting is missing
func (s Settings) StringD(key string, defaultValue string) (string, error) {
	value, err := s.String(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// MustStringD ...
func (s Settings) MustStringD(key string, defaultValue string) string {
	value, err := s.StringD(key, defaultValue)
	if err != nil {
		panic(err)
	}
	return value
}

// Bool returns a boolean setting
func (s Settings) Bool(key string) (bool, error) {
	value, ok := s.Get(key)
	if !ok {
		return false, NoSuchKeyError(key)
	}
	switch v := value.(type) {
	case bool:
		return v, nil
	case int:
		if v == 1 {
			return true, nil
		}
		if v == 0 {
			return false, nil
		}
		return false, InvalidTypeError{key, "bool", value}
	case string:
		if v == "1" || strings.EqualFold(v, "true") || strings.EqualFold(v, "t") {
			return true, nil
		}
		if v == "0" || strings.EqualFold(v, "false") || strings.EqualFold(v, "f") {
			return false, nil
		}
		return false, InvalidTypeError{key, "bool", value}
	default:
		return false, InvalidTypeError{key, "bool", value}
	}
}

// MustBool ...
func (s Settings) MustBool(key string) bool {
	value, err := s.Bool(key)
	if err != nil {
		panic(err)
	}
	return value
}

// BoolD returns a string setting value, with a default value if the setting is missing
func (s Settings) BoolD(key string, defaultValue bool) (bool, error) {
	value, err := s.Bool(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// MustBoolD ...
func (s Settings) MustBoolD(key string, defaultValue bool) bool {
	value, err := s.BoolD(key, defaultValue)
	if err != nil {
		panic(err)
	}
	return value
}

// Int returns a int setting value
func (s Settings) Int(key string) (int, error) {
	value, ok := s.Get(key)
	if !ok {
		return 0, NoSuchKeyError(key)
	}
	switch v := value.(type) {
	case int:
		return v, nil
	case string:
		i, err := strconv.Atoi(v)
		if err != nil {
			return 0, InvalidTypeError{key, "int", value}
		}
		return i, nil
	default:
		return 0, InvalidTypeError{key, "int", value}
	}
}

// MustInt ...
func (s Settings) MustInt(key string) int {
	value, err := s.Int(key)
	if err != nil {
		panic(err)
	}
	return value
}

// IntD returns a int setting value, with a default value if the setting is missing
func (s Settings) IntD(key string, defaultValue int) (int, error) {
	value, err := s.Int(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// MustIntD ...
func (s Settings) MustIntD(key string, defaultValue int) int {
	value, err := s.IntD(key, defaultValue)
	if err != nil {
		panic(err)
	}
	return value
}

// Float returns a float64 setting value
func (s Settings) Float(key string) (float64, error) {
	value, ok := s.Get(key)
	if !ok {
		return 0, NoSuchKeyError(key)
	}
	switch v := value.(type) {
	case int:
		return float64(v), nil
	case float32:
		return float64(v), nil
	case float64:
		return v, nil
	case string:
		f, err := strconv.ParseFloat(v, 64)
		if err != nil {
			return 0, InvalidTypeError{key, "float", value}
		}
		return f, nil
	default:
		return 0, InvalidTypeError{key, "float", value}
	}
}

// MustFloat ...
func (s Settings) MustFloat(key string) float64 {
	value, err := s.Float(key)
	if err != nil {
		panic(err)
	}
	return value
}

// FloatD returns a float64 setting value, with a default value if the setting is missing
func (s Settings) FloatD(key string, defaultValue float64) (float64, error) {
	value, err := s.Float(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// MustFloatD ...
func (s Settings) MustFloatD(key string, defaultValue float64) float64 {
	value, err := s.FloatD(key, defaultValue)
	if err != nil {
		panic(err)
	}
	return value
}

// StringClient returns a slice of strings setting value
func (s Settings) StringSlice(key string) ([]string, error) {
	value, ok := s[key]
	if !ok {
		return nil, NoSuchKeyError(key)
	}
	sValue, ok := value.([]string)
	if !ok {
		siValue, ok := value.([]interface{})
		if !ok {
			return nil, InvalidTypeError{key, "[string]", value}
		}
		for i, value := range siValue {
			s, ok := value.(string)
			if !ok {
				return nil, InvalidTypeError{key + strconv.Itoa(i), "string", value}
			}
			sValue = append(sValue, s)
		}
	}
	return sValue, nil
}

// MustStringSlice ...
func (s Settings) MustStringSlice(key string) []string {
	value, err := s.StringSlice(key)
	if err != nil {
		panic(err)
	}
	return value
}

// StringSliceD returns a slice of strings setting value, with a default value if the setting is missing
func (s Settings) StringSliceD(key string, defaultValue []string) ([]string, error) {
	value, err := s.StringSlice(key)
	if errors.Is(err, ErrNoSuchKey) {
		return defaultValue, nil
	}
	return value, err
}

// MustStringSliceD ...
func (s Settings) MustStringSliceD(key string, defaultValue []string) []string {
	value, err := s.StringSliceD(key, defaultValue)
	if err != nil {
		panic(err)
	}
	return value
}
